"""Basic addressable area data state and store."""
from dataclasses import dataclass
from functools import cached_property
from typing import Dict, List, Optional, Set, Union

from opentrons_shared_data.robot.types import RobotType, RobotDefinition
from opentrons_shared_data.deck.types import (
    DeckDefinitionV5,
    SlotDefV3,
    CutoutFixture,
)

from opentrons.types import Point, DeckSlotName

from ..commands import (
    Command,
    LoadLabwareResult,
    LoadModuleResult,
    MoveLabwareResult,
    MoveToAddressableAreaResult,
    MoveToAddressableAreaForDropTipResult,
)
from ..errors import (
    IncompatibleAddressableAreaError,
    AreaNotInDeckConfigurationError,
    SlotDoesNotExistError,
    AddressableAreaDoesNotExistError,
    CutoutDoesNotExistError,
)
from ..resources import deck_configuration_provider
from ..types import (
    DeckSlotLocation,
    AddressableAreaLocation,
    AddressableArea,
    PotentialCutoutFixture,
    DeckConfigurationType,
    Dimensions,
)
from ..actions import (
    Action,
    SucceedCommandAction,
    SetDeckConfigurationAction,
    AddAddressableAreaAction,
)
from .config import Config
from ._abstract_store import HasState, HandlesActions


@dataclass
class AddressableAreaState:
    """State of all loaded addressable area resources."""

    loaded_addressable_areas_by_name: Dict[str, AddressableArea]
    """The addressable areas that have been loaded so far.

    When `use_simulated_deck_config` is `False`, these are the addressable areas that the
    deck configuration provided.

    When `use_simulated_deck_config` is `True`, these are the addressable areas that have been
    referenced by the protocol so far.
    """

    potential_cutout_fixtures_by_cutout_id: Dict[str, Set[PotentialCutoutFixture]]

    deck_definition: DeckDefinitionV5

    deck_configuration: Optional[DeckConfigurationType]
    """The host robot's full deck configuration.

    If `use_simulated_deck_config` is `True`, this is meaningless and this value is undefined.
    In practice it will probably be `None` or `[]`.

    If `use_simulated_deck_config` is `False`, this will be non-`None`.
    """

    robot_type: RobotType

    use_simulated_deck_config: bool
    """See `Config.use_simulated_deck_config`."""

    """Information about the current robot model."""
    robot_definition: RobotDefinition


_OT2_ORDERED_SLOTS = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"]
_FLEX_ORDERED_SLOTS = [
    "D1",
    "D2",
    "D3",
    "C1",
    "C2",
    "C3",
    "B1",
    "B2",
    "B3",
    "A1",
    "A2",
    "A3",
]
_FLEX_ORDERED_STAGING_SLOTS = ["D4", "C4", "B4", "A4"]


def _get_conflicting_addressable_areas_error_string(
    potential_cutout_fixtures: Set[PotentialCutoutFixture],
    loaded_addressable_areas: Dict[str, AddressableArea],
    deck_definition: DeckDefinitionV5,
) -> str:
    loaded_areas_on_cutout = set()
    for fixture in potential_cutout_fixtures:
        loaded_areas_on_cutout.update(
            deck_configuration_provider.get_provided_addressable_area_names(
                fixture.cutout_fixture_id,
                fixture.cutout_id,
                deck_definition,
            )
        )
    loaded_areas_on_cutout.intersection_update(loaded_addressable_areas)
    display_names = {
        loaded_addressable_areas[area].display_name for area in loaded_areas_on_cutout
    }
    return ", ".join(display_names)


# This is a temporary shim while Protocol Engine's conflict-checking code
# can only take deck slots as input.
# Long-term solution: Check for conflicts based on bounding boxes, not slot adjacencies.
# Shorter-term: Change the conflict-checking code to take cutouts instead of deck slots.
CUTOUT_TO_DECK_SLOT_MAP: Dict[str, DeckSlotName] = {
    # OT-2
    "cutout1": DeckSlotName.SLOT_1,
    "cutout2": DeckSlotName.SLOT_2,
    "cutout3": DeckSlotName.SLOT_3,
    "cutout4": DeckSlotName.SLOT_4,
    "cutout5": DeckSlotName.SLOT_5,
    "cutout6": DeckSlotName.SLOT_6,
    "cutout7": DeckSlotName.SLOT_7,
    "cutout8": DeckSlotName.SLOT_8,
    "cutout9": DeckSlotName.SLOT_9,
    "cutout10": DeckSlotName.SLOT_10,
    "cutout11": DeckSlotName.SLOT_11,
    "cutout12": DeckSlotName.FIXED_TRASH,
    # Flex
    "cutoutA1": DeckSlotName.SLOT_A1,
    "cutoutA2": DeckSlotName.SLOT_A2,
    "cutoutA3": DeckSlotName.SLOT_A3,
    "cutoutB1": DeckSlotName.SLOT_B1,
    "cutoutB2": DeckSlotName.SLOT_B2,
    "cutoutB3": DeckSlotName.SLOT_B3,
    "cutoutC1": DeckSlotName.SLOT_C1,
    "cutoutC2": DeckSlotName.SLOT_C2,
    "cutoutC3": DeckSlotName.SLOT_C3,
    "cutoutD1": DeckSlotName.SLOT_D1,
    "cutoutD2": DeckSlotName.SLOT_D2,
    "cutoutD3": DeckSlotName.SLOT_D3,
}
DECK_SLOT_TO_CUTOUT_MAP = {
    deck_slot: cutout for cutout, deck_slot in CUTOUT_TO_DECK_SLOT_MAP.items()
}


class AddressableAreaStore(HasState[AddressableAreaState], HandlesActions):
    """Addressable area state container."""

    _state: AddressableAreaState

    def __init__(
        self,
        deck_configuration: DeckConfigurationType,
        config: Config,
        deck_definition: DeckDefinitionV5,
        robot_definition: RobotDefinition,
    ) -> None:
        """Initialize an addressable area store and its state."""
        if config.use_simulated_deck_config:
            loaded_addressable_areas_by_name = {}
        else:
            loaded_addressable_areas_by_name = (
                self._get_addressable_areas_from_deck_configuration(
                    deck_configuration,
                    deck_definition,
                )
            )

        self._state = AddressableAreaState(
            deck_configuration=deck_configuration,
            loaded_addressable_areas_by_name=loaded_addressable_areas_by_name,
            potential_cutout_fixtures_by_cutout_id={},
            deck_definition=deck_definition,
            robot_type=config.robot_type,
            use_simulated_deck_config=config.use_simulated_deck_config,
            robot_definition=robot_definition,
        )

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        if isinstance(action, SucceedCommandAction):
            self._handle_command(action.command)
        elif isinstance(action, AddAddressableAreaAction):
            self._check_location_is_addressable_area(action.addressable_area)
        elif isinstance(action, SetDeckConfigurationAction):
            current_state = self._state
            if (
                action.deck_configuration is not None
                and not self._state.use_simulated_deck_config
            ):
                self._state.deck_configuration = action.deck_configuration
                self._state.loaded_addressable_areas_by_name = (
                    self._get_addressable_areas_from_deck_configuration(
                        deck_config=action.deck_configuration,
                        deck_definition=current_state.deck_definition,
                    )
                )

    def _handle_command(self, command: Command) -> None:
        """Modify state in reaction to a command."""
        if isinstance(command.result, LoadLabwareResult):
            location = command.params.location
            if isinstance(location, (DeckSlotLocation, AddressableAreaLocation)):
                self._check_location_is_addressable_area(location)

        elif isinstance(command.result, MoveLabwareResult):
            location = command.params.newLocation
            if isinstance(location, (DeckSlotLocation, AddressableAreaLocation)):
                self._check_location_is_addressable_area(location)

        elif isinstance(command.result, LoadModuleResult):
            self._check_location_is_addressable_area(command.params.location)

        elif isinstance(
            command.result,
            (MoveToAddressableAreaResult, MoveToAddressableAreaForDropTipResult),
        ):
            addressable_area_name = command.params.addressableAreaName
            self._check_location_is_addressable_area(addressable_area_name)

    @staticmethod
    def _get_addressable_areas_from_deck_configuration(
        deck_config: DeckConfigurationType, deck_definition: DeckDefinitionV5
    ) -> Dict[str, AddressableArea]:
        """Return all addressable areas provided by the given deck configuration."""
        addressable_areas = []
        for cutout_id, cutout_fixture_id, opentrons_module_serial_number in deck_config:
            provided_addressable_areas = (
                deck_configuration_provider.get_provided_addressable_area_names(
                    cutout_fixture_id, cutout_id, deck_definition
                )
            )
            cutout_position = deck_configuration_provider.get_cutout_position(
                cutout_id, deck_definition
            )
            base_slot = CUTOUT_TO_DECK_SLOT_MAP[cutout_id]
            for addressable_area_name in provided_addressable_areas:
                addressable_areas.append(
                    deck_configuration_provider.get_addressable_area_from_name(
                        addressable_area_name=addressable_area_name,
                        cutout_position=cutout_position,
                        base_slot=base_slot,
                        deck_definition=deck_definition,
                    )
                )
        return {area.area_name: area for area in addressable_areas}

    def _check_location_is_addressable_area(
        self, location: Union[DeckSlotLocation, AddressableAreaLocation, str]
    ) -> None:
        if isinstance(location, DeckSlotLocation):
            addressable_area_name = location.slotName.id
        elif isinstance(location, AddressableAreaLocation):
            addressable_area_name = location.addressableAreaName
        else:
            addressable_area_name = location

        if addressable_area_name not in self._state.loaded_addressable_areas_by_name:
            cutout_id = self._validate_addressable_area_for_simulation(
                addressable_area_name
            )

            cutout_position = deck_configuration_provider.get_cutout_position(
                cutout_id, self._state.deck_definition
            )
            base_slot = CUTOUT_TO_DECK_SLOT_MAP[cutout_id]
            addressable_area = (
                deck_configuration_provider.get_addressable_area_from_name(
                    addressable_area_name=addressable_area_name,
                    cutout_position=cutout_position,
                    base_slot=base_slot,
                    deck_definition=self._state.deck_definition,
                )
            )
            self._state.loaded_addressable_areas_by_name[
                addressable_area.area_name
            ] = addressable_area

    def _validate_addressable_area_for_simulation(
        self, addressable_area_name: str
    ) -> str:
        """Given an addressable area name, validate it can exist on the deck and return cutout id associated with it."""
        (
            cutout_id,
            potential_fixtures,
        ) = deck_configuration_provider.get_potential_cutout_fixtures(
            addressable_area_name, self._state.deck_definition
        )

        if cutout_id in self._state.potential_cutout_fixtures_by_cutout_id:
            # Get the existing potential cutout fixtures for the addressable area already loaded on this cutout
            existing_potential_fixtures = (
                self._state.potential_cutout_fixtures_by_cutout_id[cutout_id]
            )
            # Get common cutout fixture that supplies existing addressable areas and the one being loaded
            remaining_fixtures = existing_potential_fixtures.intersection(
                set(potential_fixtures)
            )

            self._state.potential_cutout_fixtures_by_cutout_id[
                cutout_id
            ] = remaining_fixtures
        else:
            self._state.potential_cutout_fixtures_by_cutout_id[cutout_id] = set(
                potential_fixtures
            )

        return cutout_id


class AddressableAreaView(HasState[AddressableAreaState]):
    """Read-only addressable area state view."""

    _state: AddressableAreaState

    def __init__(self, state: AddressableAreaState) -> None:
        """Initialize the computed view of addressable area state.

        Arguments:
            state: Addressable area state dataclass used for all calculations.
        """
        self._state = state

    @cached_property
    def deck_extents(self) -> Point:
        """The maximum space on the deck."""
        extents = self._state.robot_definition["extents"]
        return Point(x=extents[0], y=extents[1], z=extents[2])

    @cached_property
    def mount_offsets(self) -> Dict[str, Point]:
        """The left and right mount offsets of the robot."""
        left_offset = self.state.robot_definition["mountOffsets"]["left"]
        right_offset = self.state.robot_definition["mountOffsets"]["right"]
        return {
            "left": Point(x=left_offset[0], y=left_offset[1], z=left_offset[2]),
            "right": Point(x=right_offset[0], y=right_offset[1], z=right_offset[2]),
        }

    @cached_property
    def padding_offsets(self) -> Dict[str, float]:
        """The padding offsets to be applied to the deck extents of the robot."""
        rear_offset = self.state.robot_definition["paddingOffsets"]["rear"]
        front_offset = self.state.robot_definition["paddingOffsets"]["front"]
        left_side_offset = self.state.robot_definition["paddingOffsets"]["leftSide"]
        right_side_offset = self.state.robot_definition["paddingOffsets"]["rightSide"]
        return {
            "rear": rear_offset,
            "front": front_offset,
            "left_side": left_side_offset,
            "right_side": right_side_offset,
        }

    def get_addressable_area(self, addressable_area_name: str) -> AddressableArea:
        """Get addressable area."""
        if not self._state.use_simulated_deck_config:
            return self._get_loaded_addressable_area(addressable_area_name)
        else:
            return self._get_addressable_area_from_deck_data(
                addressable_area_name=addressable_area_name,
                do_compatibility_check=True,
            )

    def get_all(self) -> List[str]:
        """Get a list of all loaded addressable area names."""
        return list(self._state.loaded_addressable_areas_by_name)

    def get_all_cutout_fixtures(self) -> Optional[List[str]]:
        """Get the names of all fixtures present in the host robot's deck configuration.

        If `use_simulated_deck_config` is `True` (see `Config`), we don't have a
        meaningful concrete layout of fixtures, so this will return `None`.
        """
        if self._state.use_simulated_deck_config:
            return None
        else:
            assert self._state.deck_configuration is not None
            return [
                cutout_fixture_id
                for _, cutout_fixture_id, _serial in self._state.deck_configuration
            ]

    def _get_loaded_addressable_area(
        self, addressable_area_name: str
    ) -> AddressableArea:
        """Get an addressable area that has been loaded into state. Will raise error if it does not exist."""
        try:
            return self._state.loaded_addressable_areas_by_name[addressable_area_name]
        except KeyError:
            raise AreaNotInDeckConfigurationError(
                f"{addressable_area_name} not provided by deck configuration."
            )

    def _check_if_area_is_compatible_with_potential_fixtures(
        self,
        area_name: str,
        cutout_id: str,
        potential_fixtures: Set[PotentialCutoutFixture],
    ) -> None:
        if cutout_id in self._state.potential_cutout_fixtures_by_cutout_id:
            if not self._state.potential_cutout_fixtures_by_cutout_id[
                cutout_id
            ].intersection(potential_fixtures):
                loaded_areas_on_cutout = (
                    _get_conflicting_addressable_areas_error_string(
                        self._state.potential_cutout_fixtures_by_cutout_id[cutout_id],
                        self._state.loaded_addressable_areas_by_name,
                        self.state.deck_definition,
                    )
                )
                area_display_name = (
                    deck_configuration_provider.get_addressable_area_display_name(
                        area_name, self.state.deck_definition
                    )
                )
                raise IncompatibleAddressableAreaError(
                    f"Cannot use {area_display_name}, not compatible with one or more of"
                    f" the following fixtures: {loaded_areas_on_cutout}"
                )

    def _get_addressable_area_from_deck_data(
        self,
        addressable_area_name: str,
        do_compatibility_check: bool,
    ) -> AddressableArea:
        """Get an addressable area that may not have been already loaded for a simulated run.

        Since this may be the first time this addressable area has been called, and it might not exist in the store
        yet (and if not won't until the result completes), we have to check if it is theoretically possible and then
        get the area data from the deck configuration provider.
        """
        if addressable_area_name in self._state.loaded_addressable_areas_by_name:
            return self._state.loaded_addressable_areas_by_name[addressable_area_name]

        (
            cutout_id,
            potential_fixtures,
        ) = deck_configuration_provider.get_potential_cutout_fixtures(
            addressable_area_name, self._state.deck_definition
        )

        if do_compatibility_check:
            self._check_if_area_is_compatible_with_potential_fixtures(
                addressable_area_name, cutout_id, potential_fixtures
            )

        cutout_position = deck_configuration_provider.get_cutout_position(
            cutout_id, self._state.deck_definition
        )
        base_slot = CUTOUT_TO_DECK_SLOT_MAP[cutout_id]
        return deck_configuration_provider.get_addressable_area_from_name(
            addressable_area_name=addressable_area_name,
            cutout_position=cutout_position,
            base_slot=base_slot,
            deck_definition=self._state.deck_definition,
        )

    def get_addressable_area_base_slot(
        self, addressable_area_name: str
    ) -> DeckSlotName:
        """Get the base slot the addressable area is associated with."""
        addressable_area = self.get_addressable_area(addressable_area_name)
        return addressable_area.base_slot

    def get_addressable_area_position(
        self,
        addressable_area_name: str,
        do_compatibility_check: bool = True,
    ) -> Point:
        """Get the position of an addressable area.

        This does not require the addressable area to be in the deck configuration.
        This is primarily used to support legacy fixed trash labware without
        modifying the deck layout to remove the similar, but functionally different,
        trashBinAdapter cutout fixture.

        Besides that instance, for movement purposes, this should only be called for
        areas that have been pre-validated, otherwise there could be the risk of collision.
        """
        addressable_area = self._get_addressable_area_from_deck_data(
            addressable_area_name=addressable_area_name,
            do_compatibility_check=False,  # This should probably not default to false
        )
        position = addressable_area.position
        return Point(x=position.x, y=position.y, z=position.z)

    def get_addressable_area_offsets_from_cutout(
        self,
        addressable_area_name: str,
    ) -> Point:
        """Get the offset form cutout fixture of an addressable area."""
        for addressable_area in self.state.deck_definition["locations"][
            "addressableAreas"
        ]:
            if addressable_area["id"] == addressable_area_name:
                area_offset = addressable_area["offsetFromCutoutFixture"]
                position = Point(
                    x=area_offset[0],
                    y=area_offset[1],
                    z=area_offset[2],
                )
                return Point(x=position.x, y=position.y, z=position.z)
        raise ValueError(
            f"No matching addressable area named {addressable_area_name} identified."
        )

    def get_addressable_area_bounding_box(
        self,
        addressable_area_name: str,
        do_compatibility_check: bool = True,
    ) -> Dimensions:
        """Get the bounding box of an addressable area.

        This does not require the addressable area to be in the deck configuration.
        For movement purposes, this should only be called for
        areas that have been pre-validated, otherwise there could be the risk of collision.
        """
        addressable_area = self._get_addressable_area_from_deck_data(
            addressable_area_name=addressable_area_name,
            do_compatibility_check=do_compatibility_check,
        )
        return addressable_area.bounding_box

    def get_addressable_area_move_to_location(
        self, addressable_area_name: str
    ) -> Point:
        """Get the move-to position (top center) for an addressable area."""
        addressable_area = self.get_addressable_area(addressable_area_name)
        position = addressable_area.position
        bounding_box = addressable_area.bounding_box
        return Point(
            x=position.x + bounding_box.x / 2,
            y=position.y + bounding_box.y / 2,
            z=position.z + bounding_box.z,
        )

    def get_addressable_area_center(self, addressable_area_name: str) -> Point:
        """Get the (x, y, z) position of the center of the area."""
        addressable_area = self.get_addressable_area(addressable_area_name)
        position = addressable_area.position
        bounding_box = addressable_area.bounding_box
        return Point(
            x=position.x + bounding_box.x / 2,
            y=position.y + bounding_box.y / 2,
            z=position.z,
        )

    def get_cutout_id_by_deck_slot_name(self, slot_name: DeckSlotName) -> str:
        """Get the Cutout ID of a given Deck Slot by Deck Slot Name."""
        return DECK_SLOT_TO_CUTOUT_MAP[slot_name]

    def get_fixture_by_deck_slot_name(
        self, slot_name: DeckSlotName
    ) -> Optional[CutoutFixture]:
        """Get the Cutout Fixture currently loaded where a specific Deck Slot would be."""
        deck_config = self.state.deck_configuration
        if deck_config:
            slot_cutout_id = DECK_SLOT_TO_CUTOUT_MAP[slot_name]
            slot_cutout_fixture = None
            # This will only ever be one under current assumptions
            for (
                cutout_id,
                cutout_fixture_id,
                opentrons_module_serial_number,
            ) in deck_config:
                if cutout_id == slot_cutout_id:
                    slot_cutout_fixture = (
                        deck_configuration_provider.get_cutout_fixture(
                            cutout_fixture_id, self.state.deck_definition
                        )
                    )
                    return slot_cutout_fixture
            if slot_cutout_fixture is None:
                # If this happens, it's a bug. Either DECK_SLOT_TO_CUTOUT_MAP
                # is missing an entry for the slot, or the deck configuration is missing
                # an entry for the cutout.
                raise CutoutDoesNotExistError(
                    f"No Cutout was found in the Deck that matched provided slot {slot_name}."
                )
        return None

    def get_fixture_height(self, cutout_fixture_name: str) -> float:
        """Get the z height of a cutout fixture."""
        cutout_fixture = deck_configuration_provider.get_cutout_fixture(
            cutout_fixture_name, self._state.deck_definition
        )
        return cutout_fixture["height"]

    def get_fixture_serial_from_deck_configuration_by_deck_slot(
        self, slot_name: DeckSlotName
    ) -> Optional[str]:
        """Get the serial number provided by the deck configuration for a Fixture at a given location."""
        deck_config = self.state.deck_configuration
        if deck_config:
            slot_cutout_id = DECK_SLOT_TO_CUTOUT_MAP[slot_name]
            # This will only ever be one under current assumptions
            for (
                cutout_id,
                cutout_fixture_id,
                opentrons_module_serial_number,
            ) in deck_config:
                if cutout_id == slot_cutout_id:
                    return opentrons_module_serial_number
        return None

    def get_slot_definition(self, slot_id: str) -> SlotDefV3:
        """Get the definition of a slot in the deck.

        This does not require that the slot exist in deck configuration.
        """
        try:
            addressable_area = self._get_addressable_area_from_deck_data(
                addressable_area_name=slot_id,
                do_compatibility_check=True,  # From the description of get_slot_definition, this might have to be False.
            )
        except AddressableAreaDoesNotExistError:
            raise SlotDoesNotExistError(
                f"Slot ID {slot_id} does not exist in deck {self._state.deck_definition['otId']}"
            )
        position = addressable_area.position
        bounding_box = addressable_area.bounding_box
        return {
            "id": addressable_area.area_name,
            "position": [position.x, position.y, position.z],
            "boundingBox": {
                "xDimension": bounding_box.x,
                "yDimension": bounding_box.y,
                "zDimension": bounding_box.z,
            },
            "displayName": addressable_area.display_name,
            "compatibleModuleTypes": addressable_area.compatible_module_types,
        }

    def get_deck_slot_definitions(self) -> Dict[str, SlotDefV3]:
        """Get all standard slot definitions available in the deck definition."""
        if self._state.robot_type == "OT-2 Standard":
            slots = _OT2_ORDERED_SLOTS
        else:
            slots = _FLEX_ORDERED_SLOTS
        return {slot_name: self.get_slot_definition(slot_name) for slot_name in slots}

    def get_staging_slot_definitions(self) -> Dict[str, SlotDefV3]:
        """Get all staging slot definitions available in the deck definition."""
        if self._state.robot_type == "OT-3 Standard":
            return {
                slot_name: self.get_slot_definition(slot_name)
                for slot_name in _FLEX_ORDERED_STAGING_SLOTS
            }
        else:
            return {}

    def raise_if_area_not_in_deck_configuration(
        self, addressable_area_name: str
    ) -> None:
        """Raise error if an addressable area is not compatible with or in the deck configuration.

        For simulated runs/analysis, this will raise if the given addressable area is not compatible with other
        previously referenced addressable areas, for example if a movable trash in A1 is in state, referencing the
        deck slot A1 will raise since those two can't exist in any deck configuration combination.

        For an on robot run, it will check if it is in the robot's deck configuration, if not it will raise an error.
        """
        if self._state.use_simulated_deck_config:
            (
                cutout_id,
                potential_fixtures,
            ) = deck_configuration_provider.get_potential_cutout_fixtures(
                addressable_area_name, self._state.deck_definition
            )

            self._check_if_area_is_compatible_with_potential_fixtures(
                addressable_area_name, cutout_id, potential_fixtures
            )
        else:
            if (
                addressable_area_name
                not in self._state.loaded_addressable_areas_by_name
            ):
                raise AreaNotInDeckConfigurationError(
                    f"{addressable_area_name} not provided by deck configuration."
                )



from typing import Dict, List, Optional, Any, Sequence, Iterator, Tuple, cast
from dataclasses import dataclass
from collections import OrderedDict
from enum import Enum
from itertools import chain

from opentrons.hardware_control.types import CriticalPoint
from opentrons.types import Point
from opentrons_shared_data.pipette.pipette_definition import (
    PipetteGeometryDefinition,
    PipetteRowDefinition,
    ValidNozzleMaps,
)
from opentrons_shared_data.errors import ErrorCodes, GeneralError, PythonException

MAXIMUM_NOZZLE_COUNT = 24


def _nozzle_names_by_row(rows: List[PipetteRowDefinition]) -> Iterator[str]:
    for row in rows:
        for nozzle in row.ordered_nozzles:
            yield nozzle


def _row_or_col_index_for_nozzle(
    row_or_col: "OrderedDict[str, List[str]]", nozzle: str
) -> int:
    for index, row_or_col_contents in enumerate(row_or_col.values()):
        if nozzle in row_or_col_contents:
            return index
    raise KeyError(nozzle)


def _row_col_indices_for_nozzle(
    rows: "OrderedDict[str, List[str]]",
    cols: "OrderedDict[str, List[str]]",
    nozzle: str,
) -> Tuple[int, int]:
    return _row_or_col_index_for_nozzle(rows, nozzle), _row_or_col_index_for_nozzle(
        cols, nozzle
    )


class NozzleConfigurationType(Enum):
    """
    Nozzle Configuration Type.

    Represents the current nozzle
    configuration stored in NozzleMap
    """

    COLUMN = "COLUMN"
    ROW = "ROW"
    SINGLE = "SINGLE"
    FULL = "FULL"
    SUBRECT = "SUBRECT"

    @classmethod
    def determine_nozzle_configuration(
        cls,
        physical_rows: "OrderedDict[str, List[str]]",
        current_rows: "OrderedDict[str, List[str]]",
        physical_cols: "OrderedDict[str, List[str]]",
        current_cols: "OrderedDict[str, List[str]]",
    ) -> "NozzleConfigurationType":
        """
        Determine the nozzle configuration based on the starting and
        ending nozzle.
        """
        if physical_rows == current_rows and physical_cols == current_cols:
            return NozzleConfigurationType.FULL
        if len(current_rows) == 1 and len(current_cols) == 1:
            return NozzleConfigurationType.SINGLE
        if len(current_rows) == 1:
            return NozzleConfigurationType.ROW
        if len(current_cols) == 1:
            return NozzleConfigurationType.COLUMN
        return NozzleConfigurationType.SUBRECT


@dataclass
class NozzleMap:
    """
    A NozzleMap instance represents a specific configuration of active nozzles on a pipette.

    It exposes properties of the configuration like the configuration's front-right, front-left,
    back-left and starting nozzles as well as a map of all the nozzles active in the configuration.

    Because NozzleMaps represent configurations directly, the properties of the NozzleMap may not
    match the properties of the physical pipette. For instance, a NozzleMap for a single channel
    configuration of an 8-channel pipette - say, A1 only - will have its front left, front right,
    and active channels all be A1, while the physical configuration would have the front right
    channel be H1.
    """

    starting_nozzle: str
    #: The nozzle that automated operations that count nozzles should start at
    # these are really ordered dicts but you can't say that even in quotes because pydantic needs to
    # evaluate them to generate serdes code so please only use ordered dicts here
    map_store: Dict[str, Point]
    #: A map of all of the nozzles active in this configuration
    valid_map_key: str
    #: A key indicating which valid nozzle map from the pipette definition represents this configuration
    rows: Dict[str, List[str]]
    #: A map of all the rows active in this configuration
    columns: Dict[str, List[str]]
    #: A map of all the columns active in this configuration
    configuration: NozzleConfigurationType
    #: The kind of configuration this is

    full_instrument_map_store: Dict[str, Point]
    #: A map of all of the nozzles of an instrument
    full_instrument_rows: Dict[str, List[str]]
    #: A map of all the rows of an instrument

    def __str__(self) -> str:
        return f"back_left_nozzle: {self.back_left} front_right_nozzle: {self.front_right} configuration: {self.configuration}"

    @property
    def back_left(self) -> str:
        """The backest, leftest (i.e. back if it's a column, left if it's a row) nozzle of the configuration.

        Note: This is the value relevant for this particular configuration, and it may not represent the back left nozzle
        of the underlying physical pipette. For instance, the back-left nozzle of a configuration representing nozzles
        D7 to H12 of a 96-channel pipette is D7, which is not the back-left nozzle of the physical pipette (A1).
        """
        return next(iter(self.rows.values()))[0]

    @property
    def front_right(self) -> str:
        """The frontest, rightest (i.e. front if it's a column, right if it's a row) nozzle of the configuration.

        Note: This is the value relevant for this configuration, not the physical pipette. See the note on back_left.
        """
        return next(reversed(list(self.rows.values())))[-1]

    @property
    def full_instrument_back_left(self) -> str:
        """The backest, leftest (i.e. back if it's a column, left if it's a row) nozzle of the full instrument.

        Note: This value represents the back left nozzle of the underlying physical pipette. For instance,
        the back-left nozzle of a 96-Channel pipette is A1.
        """
        return next(iter(self.full_instrument_rows.values()))[0]

    @property
    def full_instrument_front_right(self) -> str:
        """The frontest, rightest (i.e. front if it's a column, right if it's a row) nozzle of the full instrument.

        Note: This value represents the front right nozzle of the physical pipette. See the note on full_instrument_back_left.
        """
        return next(reversed(list(self.full_instrument_rows.values())))[-1]

    @property
    def starting_nozzle_offset(self) -> Point:
        """The position of the starting nozzle."""
        return self.map_store[self.starting_nozzle]

    @property
    def xy_center_offset(self) -> Point:
        """The position of the geometrical center of all nozzles in the configuration.

        Note: This is the value relevant for this configuration, not the physical pipette. See the note on back_left.
        """
        difference = self.map_store[self.front_right] - self.map_store[self.back_left]
        return self.map_store[self.back_left] + Point(
            difference[0] / 2, difference[1] / 2, 0
        )

    @property
    def instrument_xy_center_offset(self) -> Point:
        """The position of the geometrical center of all nozzles for the entire instrument.

        Note: This the value reflects the center of the maximum number of nozzles of the physical pipette.
        This would be the same as a full configuration.
        """
        difference = (
            self.full_instrument_map_store[self.full_instrument_front_right]
            - self.full_instrument_map_store[self.full_instrument_back_left]
        )
        return self.full_instrument_map_store[self.full_instrument_back_left] + Point(
            difference[0] / 2, difference[1] / 2, 0
        )

    @property
    def y_center_offset(self) -> Point:
        """The position in the center of the primary column of the map."""
        front_left = next(reversed(list(self.rows.values())))[0]
        difference = self.map_store[front_left] - self.map_store[self.back_left]
        return self.map_store[self.back_left] + Point(0, difference[1] / 2, 0)

    @property
    def front_nozzle_offset(self) -> Point:
        """The offset for the front_left nozzle."""
        # front left-most nozzle of the 96 channel in a given configuration
        # and front nozzle of the 8 channel
        front_left = next(iter(self.columns.values()))[-1]
        return self.map_store[front_left]

    @property
    def front_right_nozzle_offset(self) -> Point:
        """The offset for the front_right nozzle."""
        # Front-right-most nozzle of the 96 channel in a given configuration
        # and Front-most nozzle of the 8-channel
        return self.map_store[self.front_right]

    @property
    def back_left_nozzle_offset(self) -> Point:
        """The offset for the back_left nozzle."""
        # Back-left-most nozzle of the 96-channel in a given configuration
        # and back-most nozzle of the 8-channel
        return self.map_store[self.back_left]

    @property
    def tip_count(self) -> int:
        """The total number of active nozzles in the configuration, and thus the number of tips that will be picked up."""
        return len(self.map_store)

    @classmethod
    def build(  # noqa: C901
        cls,
        physical_nozzles: "OrderedDict[str, Point]",
        physical_rows: "OrderedDict[str, List[str]]",
        physical_columns: "OrderedDict[str, List[str]]",
        starting_nozzle: str,
        back_left_nozzle: str,
        front_right_nozzle: str,
        valid_nozzle_maps: ValidNozzleMaps,
    ) -> "NozzleMap":
        try:
            back_left_row_index, back_left_column_index = _row_col_indices_for_nozzle(
                physical_rows, physical_columns, back_left_nozzle
            )
        except KeyError as e:
            raise IncompatibleNozzleConfiguration(
                message=f"No entry for back left nozzle {e} in pipette",
                wrapping=[PythonException(e)],
            ) from e
        try:
            (
                front_right_row_index,
                front_right_column_index,
            ) = _row_col_indices_for_nozzle(
                physical_rows, physical_columns, front_right_nozzle
            )
        except KeyError as e:
            raise IncompatibleNozzleConfiguration(
                message=f"No entry for front right nozzle {e} in pipette",
                wrapping=[PythonException(e)],
            ) from e

        correct_rows_with_all_columns = list(physical_rows.items())[
            back_left_row_index : front_right_row_index + 1
        ]
        correct_rows = [
            (
                row_name,
                row_entries[back_left_column_index : front_right_column_index + 1],
            )
            for row_name, row_entries in correct_rows_with_all_columns
        ]
        rows = OrderedDict(correct_rows)
        correct_columns_with_all_rows = list(physical_columns.items())[
            back_left_column_index : front_right_column_index + 1
        ]
        correct_columns = [
            (col_name, col_entries[back_left_row_index : front_right_row_index + 1])
            for col_name, col_entries in correct_columns_with_all_rows
        ]
        columns = OrderedDict(correct_columns)

        map_store = OrderedDict(
            (nozzle, physical_nozzles[nozzle]) for nozzle in chain(*rows.values())
        )

        if (
            NozzleConfigurationType.determine_nozzle_configuration(
                physical_rows, rows, physical_columns, columns
            )
            != NozzleConfigurationType.FULL
        ):
            if len(rows) * len(columns) > MAXIMUM_NOZZLE_COUNT:
                raise IncompatibleNozzleConfiguration(
                    f"Partial Nozzle Layouts may not be configured to contain more than {MAXIMUM_NOZZLE_COUNT} channels."
                )

        validated_map_key = None
        for map_key in valid_nozzle_maps.maps.keys():
            if valid_nozzle_maps.maps[map_key] == list(map_store.keys()):
                validated_map_key = map_key
                break
        if validated_map_key is None:
            raise IncompatibleNozzleConfiguration(
                "Attempted Nozzle Configuration does not match any approved map layout for the current pipette."
            )

        return cls(
            starting_nozzle=starting_nozzle,
            map_store=map_store,
            valid_map_key=validated_map_key,
            rows=rows,
            full_instrument_map_store=physical_nozzles,
            full_instrument_rows=physical_rows,
            columns=columns,
            configuration=NozzleConfigurationType.determine_nozzle_configuration(
                physical_rows, rows, physical_columns, columns
            ),
        )


class IncompatibleNozzleConfiguration(GeneralError):
    """Error raised if nozzle configuration is incompatible with the currently loaded pipette."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[GeneralError]] = None,
    ) -> None:
        """Build a IncompatibleNozzleConfiguration error."""
        super().__init__(
            code=ErrorCodes.API_MISCONFIGURATION,
            message=message,
            detail=detail,
            wrapping=wrapping,
        )


class NozzleConfigurationManager:
    def __init__(
        self, nozzle_map: NozzleMap, valid_nozzle_maps: ValidNozzleMaps
    ) -> None:
        self._physical_nozzle_map = nozzle_map
        self._current_nozzle_configuration = nozzle_map
        self._valid_nozzle_maps = valid_nozzle_maps

    @classmethod
    def build_from_config(
        cls,
        pipette_geometry: PipetteGeometryDefinition,
        valid_nozzle_maps: ValidNozzleMaps,
    ) -> "NozzleConfigurationManager":
        sorted_nozzle_map = OrderedDict(
            (
                (k, Point(*pipette_geometry.nozzle_map[k]))
                for k in _nozzle_names_by_row(pipette_geometry.ordered_rows)
            )
        )
        sorted_rows = OrderedDict(
            (entry.key, entry.ordered_nozzles)
            for entry in pipette_geometry.ordered_rows
        )
        sorted_cols = OrderedDict(
            (entry.key, entry.ordered_nozzles)
            for entry in pipette_geometry.ordered_columns
        )
        back_left = next(iter(sorted_rows.values()))[0]
        front_right = next(reversed(list(sorted_rows.values())))[-1]
        starting_nozzle_config = NozzleMap.build(
            physical_nozzles=sorted_nozzle_map,
            physical_rows=sorted_rows,
            physical_columns=sorted_cols,
            starting_nozzle=back_left,
            back_left_nozzle=back_left,
            front_right_nozzle=front_right,
            valid_nozzle_maps=valid_nozzle_maps,
        )
        return cls(starting_nozzle_config, valid_nozzle_maps)

    @property
    def starting_nozzle_offset(self) -> Point:
        return self._current_nozzle_configuration.starting_nozzle_offset

    @property
    def current_configuration(self) -> NozzleMap:
        return self._current_nozzle_configuration

    def reset_to_default_configuration(self) -> None:
        self._current_nozzle_configuration = self._physical_nozzle_map

    def update_nozzle_configuration(
        self,
        back_left_nozzle: str,
        front_right_nozzle: str,
        starting_nozzle: Optional[str] = None,
    ) -> None:
        self._current_nozzle_configuration = NozzleMap.build(
            # these casts are because of pydantic in the protocol engine (see above)
            physical_nozzles=cast(
                "OrderedDict[str, Point]", self._physical_nozzle_map.map_store
            ),
            physical_rows=cast(
                "OrderedDict[str, List[str]]", self._physical_nozzle_map.rows
            ),
            physical_columns=cast(
                "OrderedDict[str, List[str]]", self._physical_nozzle_map.columns
            ),
            starting_nozzle=starting_nozzle or back_left_nozzle,
            back_left_nozzle=back_left_nozzle,
            front_right_nozzle=front_right_nozzle,
            valid_nozzle_maps=self._valid_nozzle_maps,
        )

    def get_tip_count(self) -> int:
        return self._current_nozzle_configuration.tip_count

    def critical_point_with_tip_length(
        self,
        cp_override: Optional[CriticalPoint],
        tip_length: float = 0.0,
    ) -> Point:
        if cp_override == CriticalPoint.INSTRUMENT_XY_CENTER:
            current_nozzle = (
                self._current_nozzle_configuration.instrument_xy_center_offset
            )
        elif cp_override == CriticalPoint.XY_CENTER:
            current_nozzle = self._current_nozzle_configuration.xy_center_offset
        elif cp_override == CriticalPoint.Y_CENTER:
            current_nozzle = self._current_nozzle_configuration.y_center_offset
        elif cp_override == CriticalPoint.FRONT_NOZZLE:
            current_nozzle = self._current_nozzle_configuration.front_nozzle_offset
        else:
            current_nozzle = self.starting_nozzle_offset
        return current_nozzle - Point(0, 0, tip_length)



""" Adapters for the :py:class:`.hardware_control.API` instances.
"""
import asyncio
import functools
from typing import Generic, TypeVar, Callable, Any, cast, Awaitable
from .protocols import AsyncioConfigurable


WrappedObj = TypeVar("WrappedObj", bound=AsyncioConfigurable, covariant=True)
WrappedReturn = TypeVar("WrappedReturn")


# TODO: BC 2020-02-25 instead of overwriting __get_attribute__ in this class
# use inspect.getmembers to iterate over appropriate members of adapted
# instance and setattr on the outer instance with the proper async resolution
# logic injected. This approach avoids requiring calls to
# object.__get_attribute__(self,...) to opt out of the overwritten
# functionality. It is more readable and protected from
# unintentional recursion.
class SynchronousAdapter(Generic[WrappedObj]):
    """A wrapper to make every call into :py:class:`.hardware_control.API`
    synchronous.

    This class expects to wrap an asynchronous object running in its own thread
    and event loop (obj._loop). Attempting to instantiate a SynchronousAdapter
    in the main thread within it's event loop will hang unless the adapted
    async object is running on its own thread and contained loop.
    In these Cases, it is often helpful to instantiate the API via the
    :py:class:`opentrons.hardware_control.ThreadManager` to ensure that
    all API coroutines are resolved in a thread/loop other than the
    main thread/loop.

    Example
    -------
    .. code-block::
    >>> import opentrons.hardware_control as hc
    >>> import opentrons.hardware_control.adapters as adapts
    >>> api = hc.API.build_hardware_simulator()
    >>> sync_api = adapts.SynchronousAdapter(api)
    >>> sync_api.home()
    """

    def __init__(self, asynchronous_instance: WrappedObj) -> None:
        """Build the SynchronousAdapter.

        :param asynchronous_instance: The asynchronous class instance to wrap
        """
        self._obj_to_adapt = asynchronous_instance

    def __repr__(self) -> str:
        return "<SynchronousAdapter>"

    @staticmethod
    def call_coroutine_sync(
        loop: asyncio.AbstractEventLoop,
        to_call: Callable[..., Awaitable[WrappedReturn]],
        *args: Any,
        **kwargs: Any,
    ) -> WrappedReturn:
        fut = cast(
            "asyncio.Future[WrappedReturn]",
            asyncio.run_coroutine_threadsafe(to_call(*args, **kwargs), loop),
        )
        return fut.result()

    def __getattribute__(self, attr_name: str) -> Any:
        """Retrieve attributes from our API and wrap coroutines"""
        # Almost every attribute retrieved from us will be for people actually
        # looking for an attribute of the hardware API, so check there first.
        obj_to_adapt = object.__getattribute__(self, "_obj_to_adapt")
        try:
            inner_attr = getattr(obj_to_adapt, attr_name)
        except AttributeError:
            # Maybe this actually was for us? Let’s find it
            return object.__getattribute__(self, attr_name)

        check = inner_attr
        if isinstance(inner_attr, functools.partial):
            # if partial func check passed in func
            check = inner_attr.func
        try:
            # if decorated func check wrapped func
            check = check.__wrapped__
        except AttributeError:
            pass
        if asyncio.iscoroutinefunction(check):
            # Return a synchronized version of the coroutine
            return functools.partial(
                object.__getattribute__(self, "call_coroutine_sync"),
                obj_to_adapt._loop,
                inner_attr,
            )
        elif asyncio.iscoroutine(check):
            # Catch awaitable properties and reify the future before returning
            fut = asyncio.run_coroutine_threadsafe(check, obj_to_adapt._loop)
            return fut.result()

        return inner_attr



from typing import Tuple

from opentrons.types import Point


def get_relative_offset(
    point: Point,
    size: Tuple[float, float, float],
    x_ratio: float,
    y_ratio: float,
    z_ratio: float,
) -> Point:
    """Gets a relative offset for a deck coordinate based on percentage of the radius of each axis."""
    x_size, y_size, z_size = size

    x_offset = x_size / 2.0 * x_ratio
    y_offset = y_size / 2.0 * y_ratio
    z_offset = z_size / 2.0 * z_ratio

    return Point(
        x=point.x + x_offset,
        y=point.y + y_offset,
        z=point.z + z_offset,
    )


def are_overlapping_rectangles(
    rectangle1: Tuple[Point, Point],
    rectangle2: Tuple[Point, Point],
) -> bool:
    """Return whether the two provided rectangles are overlapping in 2D space.

    The rectangles are assumed to be coplanar and represented by tuples of
    the back-left and front right vertices (in that order) of the respective rectangles.
    The z-coordinate of each point will be ignored.

    We determine if the rectangles overlap by comparing projections of the sides of
    the rectangles on each of the 2 axes (x & y). If the projections on each axis overlap,
    then we can conclude that the rectangles overlap.

    The projection on an axis overlaps if the distance between the first projected point
    and the last projected point is less than the sum of the lengths of the projected sides
    of the two rectangles. For example, if we have two rectangles with vertices:
    Rect1 -> BL: (x1, y1), FR: (x2, y2)
    Rect2 -> BL: (x3, y3), FR: (x4, y4)

    Then for the two rectangles to be overlapping, they should satisfy:
    max(x1, x2, x3, x4) - min(x1, x2, x3, x4) < (x2 - x1) + (x4 - x3)
    AND
    max(y1, y2, y3, y4) - min(y1, y2, y3, y4) < (y2 - y1) + (y4 - y3)
    """
    x_coordinates = [rectangle1[0].x, rectangle1[1].x, rectangle2[0].x, rectangle2[1].x]
    x_length_rect1 = abs(rectangle1[1].x - rectangle1[0].x)
    x_length_rect2 = abs(rectangle2[1].x - rectangle2[0].x)
    overlapping_in_x = (
        abs(max(x_coordinates) - min(x_coordinates)) < x_length_rect1 + x_length_rect2
    )
    y_coordinates = [rectangle1[0].y, rectangle1[1].y, rectangle2[0].y, rectangle2[1].y]
    y_length_rect1 = abs(rectangle1[1].y - rectangle1[0].y)
    y_length_rect2 = abs(rectangle2[1].y - rectangle2[0].y)
    overlapping_in_y = (
        abs(max(y_coordinates) - min(y_coordinates)) < y_length_rect1 + y_length_rect2
    )
    return overlapping_in_x and overlapping_in_y



"""Geometry state getters."""
import enum
from numpy import array, dot, double as npdouble
from numpy.typing import NDArray
from typing import Optional, List, Tuple, Union, cast, TypeVar, Dict
from dataclasses import dataclass
from functools import cached_property

from opentrons.types import Point, DeckSlotName, StagingSlotName, MountType

from opentrons_shared_data.labware.constants import WELL_NAME_PATTERN
from opentrons_shared_data.deck.types import CutoutFixture
from opentrons_shared_data.pipette import PIPETTE_X_SPAN
from opentrons_shared_data.pipette.types import ChannelCount
from opentrons.protocols.models import LabwareDefinition

from .. import errors
from ..errors import (
    LabwareNotLoadedOnLabwareError,
    LabwareNotLoadedOnModuleError,
    LabwareMovementNotAllowedError,
    OperationLocationNotInWellError,
)
from ..resources import fixture_validation, labware_validation
from ..types import (
    OFF_DECK_LOCATION,
    LoadedLabware,
    LoadedModule,
    WellLocation,
    LiquidHandlingWellLocation,
    DropTipWellLocation,
    PickUpTipWellLocation,
    WellOrigin,
    DropTipWellOrigin,
    WellOffset,
    DeckSlotLocation,
    ModuleLocation,
    OnLabwareLocation,
    LabwareLocation,
    LabwareOffsetVector,
    ModuleOffsetVector,
    ModuleOffsetData,
    CurrentWell,
    CurrentPipetteLocation,
    TipGeometry,
    LabwareMovementOffsetData,
    OnDeckLabwareLocation,
    AddressableAreaLocation,
    AddressableOffsetVector,
    StagingSlotLocation,
    LabwareOffsetLocation,
    ModuleModel,
)
from .config import Config
from .labware import LabwareView
from .wells import WellView
from .modules import ModuleView
from .pipettes import PipetteView
from .addressable_areas import AddressableAreaView
from .frustum_helpers import (
    find_volume_at_well_height,
    find_height_at_well_volume,
)


SLOT_WIDTH = 128
_PIPETTE_HOMED_POSITION_Z = (
    248.0  # Height of the bottom of the nozzle without the tip attached when homed
)


class _TipDropSection(enum.Enum):
    """Well sections to drop tips in."""

    LEFT = "left"
    RIGHT = "right"


class _GripperMoveType(enum.Enum):
    """Types of gripper movement."""

    PICK_UP_LABWARE = enum.auto()
    DROP_LABWARE = enum.auto()


@dataclass
class _AbsoluteRobotExtents:
    front_left: Dict[MountType, Point]
    back_right: Dict[MountType, Point]
    deck_extents: Point
    padding_rear: float
    padding_front: float
    padding_left_side: float
    padding_right_side: float


_LabwareLocation = TypeVar("_LabwareLocation", bound=LabwareLocation)


# TODO(mc, 2021-06-03): continue evaluation of which selectors should go here
# vs which selectors should be in LabwareView
class GeometryView:
    """Geometry computed state getters."""

    def __init__(
        self,
        config: Config,
        labware_view: LabwareView,
        well_view: WellView,
        module_view: ModuleView,
        pipette_view: PipetteView,
        addressable_area_view: AddressableAreaView,
    ) -> None:
        """Initialize a GeometryView instance."""
        self._config = config
        self._labware = labware_view
        self._wells = well_view
        self._modules = module_view
        self._pipettes = pipette_view
        self._addressable_areas = addressable_area_view
        self._last_drop_tip_location_spot: Dict[str, _TipDropSection] = {}

    @cached_property
    def absolute_deck_extents(self) -> _AbsoluteRobotExtents:
        """The absolute deck extents for a given robot deck."""
        left_offset = self._addressable_areas.mount_offsets["left"]
        right_offset = self._addressable_areas.mount_offsets["right"]

        front_left_abs = {
            MountType.LEFT: Point(left_offset.x, -1 * left_offset.y, left_offset.z),
            MountType.RIGHT: Point(right_offset.x, -1 * right_offset.y, right_offset.z),
        }
        back_right_abs = {
            MountType.LEFT: self._addressable_areas.deck_extents + left_offset,
            MountType.RIGHT: self._addressable_areas.deck_extents + right_offset,
        }
        return _AbsoluteRobotExtents(
            front_left=front_left_abs,
            back_right=back_right_abs,
            deck_extents=self._addressable_areas.deck_extents,
            padding_rear=self._addressable_areas.padding_offsets["rear"],
            padding_front=self._addressable_areas.padding_offsets["front"],
            padding_left_side=self._addressable_areas.padding_offsets["left_side"],
            padding_right_side=self._addressable_areas.padding_offsets["right_side"],
        )

    def get_labware_highest_z(self, labware_id: str) -> float:
        """Get the highest Z-point of a labware."""
        labware_data = self._labware.get(labware_id)

        return self._get_highest_z_from_labware_data(labware_data)

    def get_all_obstacle_highest_z(self) -> float:
        """Get the highest Z-point across all obstacles that the instruments need to fly over."""
        highest_labware_z = max(
            (
                self._get_highest_z_from_labware_data(lw_data)
                for lw_data in self._labware.get_all()
                if lw_data.location != OFF_DECK_LOCATION
            ),
            default=0.0,
        )

        # Fixme (spp, 2023-12-04): the overall height is not the true highest z of modules
        #  on a Flex.
        highest_module_z = max(
            (
                self._modules.get_overall_height(module.id)
                for module in self._modules.get_all()
            ),
            default=0.0,
        )

        cutout_fixture_names = self._addressable_areas.get_all_cutout_fixtures()
        if cutout_fixture_names is None:
            # We're using a simulated deck config (see `Config.use_simulated_deck_config`).
            # We only know the addressable areas referenced by the protocol, not the fixtures
            # providing them. And there is more than one possible configuration of fixtures
            # to provide them. So, we can't know what the highest fixture is. Default to 0.
            #
            # Defaulting to 0 may not be the right thing to do here.
            # For example, suppose a protocol references an addressable area that implies a tall
            # fixture must be on the deck, and then it uses long tips that wouldn't be able to
            # clear the top of that fixture. We should perhaps raise an analysis error for that,
            # but defaulting to 0 here means we won't.
            highest_fixture_z = 0.0
        else:
            highest_fixture_z = max(
                (
                    self._addressable_areas.get_fixture_height(cutout_fixture_name)
                    for cutout_fixture_name in cutout_fixture_names
                ),
                default=0.0,
            )

        return max(
            highest_labware_z,
            highest_module_z,
            highest_fixture_z,
        )

    def get_highest_z_in_slot(
        self, slot: Union[DeckSlotLocation, StagingSlotLocation]
    ) -> float:
        """Get the highest Z-point of all items stacked in the given deck slot.

        This height includes the height of any module that occupies the given slot
        even if it wasn't loaded in that slot (e.g., thermocycler).
        """
        slot_item = self.get_slot_item(slot.slotName)
        if isinstance(slot_item, LoadedModule):
            # get height of module + all labware on it
            module_id = slot_item.id
            try:
                labware_id = self._labware.get_id_by_module(module_id=module_id)
            except LabwareNotLoadedOnModuleError:
                return self._modules.get_module_highest_z(
                    module_id=module_id,
                    addressable_areas=self._addressable_areas,
                )
            else:
                return self.get_highest_z_of_labware_stack(labware_id)
        elif isinstance(slot_item, LoadedLabware):
            # get stacked heights of all labware in the slot
            return self.get_highest_z_of_labware_stack(slot_item.id)
        elif type(slot_item) is dict:
            # TODO (cb, 2024-02-05): Eventually this logic should become the responsibility of bounding box
            # conflict checking, as fixtures may not always be considered as items from slots.
            return self._addressable_areas.get_fixture_height(slot_item["id"])
        else:
            return 0

    def get_highest_z_of_labware_stack(self, labware_id: str) -> float:
        """Get the highest Z-point of the topmost labware in the stack of labware on the given labware.

        If there is no labware on the given labware, returns highest z of the given labware.
        """
        try:
            stacked_labware_id = self._labware.get_id_by_labware(labware_id)
        except LabwareNotLoadedOnLabwareError:
            return self.get_labware_highest_z(labware_id)
        return self.get_highest_z_of_labware_stack(stacked_labware_id)

    def get_min_travel_z(
        self,
        pipette_id: str,
        labware_id: str,
        location: Optional[CurrentPipetteLocation],
        minimum_z_height: Optional[float],
    ) -> float:
        """Get the minimum allowed travel height of an arc move."""
        if (
            isinstance(location, CurrentWell)
            and pipette_id == location.pipette_id
            and labware_id == location.labware_id
        ):
            min_travel_z = self.get_labware_highest_z(labware_id)
        else:
            min_travel_z = self.get_all_obstacle_highest_z()
        if minimum_z_height:
            min_travel_z = max(min_travel_z, minimum_z_height)
        return min_travel_z

    def get_labware_parent_nominal_position(self, labware_id: str) -> Point:
        """Get the position of the labware's uncalibrated parent (deck slot, module, or another labware)."""
        try:
            addressable_area_name = self.get_ancestor_slot_name(labware_id).id
        except errors.LocationIsStagingSlotError:
            addressable_area_name = self._get_staging_slot_name(labware_id)
        except errors.LocationIsLidDockSlotError:
            addressable_area_name = self._get_lid_dock_slot_name(labware_id)
        parent_pos = self._addressable_areas.get_addressable_area_position(
            addressable_area_name
        )

        offset_from_parent = self._get_offset_from_parent(
            child_definition=self._labware.get_definition(labware_id),
            parent=self._labware.get(labware_id).location,
        )

        return Point(
            parent_pos.x + offset_from_parent.x,
            parent_pos.y + offset_from_parent.y,
            parent_pos.z + offset_from_parent.z,
        )

    def _get_offset_from_parent(
        self, child_definition: LabwareDefinition, parent: LabwareLocation
    ) -> LabwareOffsetVector:
        """Gets the offset vector of a labware placed on the given location.

        - For labware on Deck Slot: returns an offset of (0, 0, 0)
        - For labware on a Module: returns the nominal offset for the labware's position
          when placed on the specified module (using slot-transformed labwareOffset
          from the module's definition with any stacking overlap).
          Does not include module calibration offset or LPC offset.
        - For labware on another labware: returns the nominal offset for the labware
          as placed on the specified labware, taking into account any offsets for labware
          on modules as well as stacking overlaps.
          Does not include module calibration offset or LPC offset.
        """
        if isinstance(parent, (AddressableAreaLocation, DeckSlotLocation)):
            return LabwareOffsetVector(x=0, y=0, z=0)
        elif isinstance(parent, ModuleLocation):
            module_id = parent.moduleId
            module_to_child = self._modules.get_nominal_offset_to_child(
                module_id=module_id, addressable_areas=self._addressable_areas
            )
            module_model = self._modules.get_connected_model(module_id)
            stacking_overlap = self._labware.get_module_overlap_offsets(
                child_definition, module_model
            )
            return LabwareOffsetVector(
                x=module_to_child.x - stacking_overlap.x,
                y=module_to_child.y - stacking_overlap.y,
                z=module_to_child.z - stacking_overlap.z,
            )
        elif isinstance(parent, OnLabwareLocation):
            on_labware = self._labware.get(parent.labwareId)
            on_labware_dimensions = self._labware.get_dimensions(
                labware_id=on_labware.id
            )
            stacking_overlap = self._labware.get_labware_overlap_offsets(
                definition=child_definition, below_labware_name=on_labware.loadName
            )
            labware_offset = LabwareOffsetVector(
                x=stacking_overlap.x,
                y=stacking_overlap.y,
                z=on_labware_dimensions.z - stacking_overlap.z,
            )
            return labware_offset + self._get_offset_from_parent(
                self._labware.get_definition(on_labware.id), on_labware.location
            )
        else:
            raise errors.LabwareNotOnDeckError(
                "Cannot access labware since it is not on the deck. "
                "Either it has been loaded off-deck or its been moved off-deck."
            )

    def _normalize_module_calibration_offset(
        self,
        module_location: DeckSlotLocation,
        offset_data: Optional[ModuleOffsetData],
    ) -> ModuleOffsetVector:
        """Normalize the module calibration offset depending on the module location."""
        if not offset_data:
            return ModuleOffsetVector(x=0, y=0, z=0)
        offset = offset_data.moduleOffsetVector
        calibrated_slot = offset_data.location.slotName
        calibrated_slot_column = self.get_slot_column(calibrated_slot)
        current_slot_column = self.get_slot_column(module_location.slotName)
        # make sure that we have valid colums since we cant have modules in the middle of the deck
        assert set([calibrated_slot_column, current_slot_column]).issubset(
            {1, 3}
        ), f"Module calibration offset is an invalid slot {calibrated_slot}"

        # Check if the module has moved from one side of the deck to the other
        if calibrated_slot_column != current_slot_column:
            # Since the module was rotated, the calibration offset vector needs to be rotated by 180 degrees along the z axis
            saved_offset: NDArray[npdouble] = array([offset.x, offset.y, offset.z])
            rotation_matrix: NDArray[npdouble] = array(
                [[-1, 0, 0], [0, -1, 0], [0, 0, 1]]
            )
            new_offset = dot(saved_offset, rotation_matrix)
            offset = ModuleOffsetVector(
                x=new_offset[0], y=new_offset[1], z=new_offset[2]
            )
        return offset

    def _get_calibrated_module_offset(
        self, location: LabwareLocation
    ) -> ModuleOffsetVector:
        """Get a labware location's underlying calibrated module offset, if it is on a module."""
        if isinstance(location, ModuleLocation):
            module_id = location.moduleId
            module_location = self._modules.get_location(module_id)
            offset_data = self._modules.get_module_calibration_offset(module_id)
            return self._normalize_module_calibration_offset(
                module_location, offset_data
            )
        elif isinstance(location, (DeckSlotLocation, AddressableAreaLocation)):
            # TODO we might want to do a check here to make sure addressable area location is a standard deck slot
            #   and raise if its not (or maybe we don't actually care since modules will never be loaded elsewhere)
            return ModuleOffsetVector(x=0, y=0, z=0)
        elif isinstance(location, OnLabwareLocation):
            labware_data = self._labware.get(location.labwareId)
            return self._get_calibrated_module_offset(labware_data.location)
        elif location == OFF_DECK_LOCATION:
            raise errors.LabwareNotOnDeckError(
                "Labware does not have a slot or module associated with it"
                " since it is no longer on the deck."
            )

    def get_labware_parent_position(self, labware_id: str) -> Point:
        """Get the calibrated position of the labware's parent slot (deck or module)."""
        parent_pos = self.get_labware_parent_nominal_position(labware_id)
        labware_data = self._labware.get(labware_id)
        cal_offset = self._get_calibrated_module_offset(labware_data.location)

        return Point(
            x=parent_pos.x + cal_offset.x,
            y=parent_pos.y + cal_offset.y,
            z=parent_pos.z + cal_offset.z,
        )

    def get_labware_origin_position(self, labware_id: str) -> Point:
        """Get the position of the labware's origin, without calibration."""
        slot_pos = self.get_labware_parent_position(labware_id)
        origin_offset = self._labware.get_definition(labware_id).cornerOffsetFromSlot

        return Point(
            x=slot_pos.x + origin_offset.x,
            y=slot_pos.y + origin_offset.y,
            z=slot_pos.z + origin_offset.z,
        )

    def get_labware_position(self, labware_id: str) -> Point:
        """Get the calibrated origin of the labware."""
        origin_pos = self.get_labware_origin_position(labware_id)
        cal_offset = self._labware.get_labware_offset_vector(labware_id)

        return Point(
            x=origin_pos.x + cal_offset.x,
            y=origin_pos.y + cal_offset.y,
            z=origin_pos.z + cal_offset.z,
        )

    WellLocations = Union[
        WellLocation, LiquidHandlingWellLocation, PickUpTipWellLocation
    ]

    def validate_well_position(
        self,
        well_location: WellLocations,
        z_offset: float,
        pipette_id: Optional[str] = None,
    ) -> None:
        """Raise exception if operation location is not within well.

        Primarily this checks if there is not enough liquid in a well to do meniscus-relative static aspiration.
        """
        if well_location.origin == WellOrigin.MENISCUS:
            assert pipette_id is not None
            lld_min_height = self._pipettes.get_current_tip_lld_settings(
                pipette_id=pipette_id
            )
            if z_offset < lld_min_height:
                if isinstance(well_location, PickUpTipWellLocation):
                    raise OperationLocationNotInWellError(
                        f"Specifying {well_location.origin} with an offset of {well_location.offset} results in an operation location that could be below the bottom of the well"
                    )
                else:
                    raise OperationLocationNotInWellError(
                        f"Specifying {well_location.origin} with an offset of {well_location.offset} and a volume offset of {well_location.volumeOffset} results in an operation location that could be below the bottom of the well"
                    )
        elif z_offset < 0:
            if isinstance(well_location, PickUpTipWellLocation):
                raise OperationLocationNotInWellError(
                    f"Specifying {well_location.origin} with an offset of {well_location.offset} results in an operation location below the bottom of the well"
                )
            else:
                raise OperationLocationNotInWellError(
                    f"Specifying {well_location.origin} with an offset of {well_location.offset} and a volume offset of {well_location.volumeOffset} results in an operation location below the bottom of the well"
                )

    def get_well_position(
        self,
        labware_id: str,
        well_name: str,
        well_location: Optional[WellLocations] = None,
        operation_volume: Optional[float] = None,
        pipette_id: Optional[str] = None,
    ) -> Point:
        """Given relative well location in a labware, get absolute position."""
        labware_pos = self.get_labware_position(labware_id)
        well_def = self._labware.get_well_definition(labware_id, well_name)
        well_depth = well_def.depth

        offset = WellOffset(x=0, y=0, z=well_depth)
        if well_location is not None:
            offset = well_location.offset
            offset_adjustment = self.get_well_offset_adjustment(
                labware_id=labware_id,
                well_name=well_name,
                well_location=well_location,
                well_depth=well_depth,
                operation_volume=operation_volume,
            )
            offset = offset.copy(update={"z": offset.z + offset_adjustment})
            self.validate_well_position(
                well_location=well_location, z_offset=offset.z, pipette_id=pipette_id
            )

        return Point(
            x=labware_pos.x + offset.x + well_def.x,
            y=labware_pos.y + offset.y + well_def.y,
            z=labware_pos.z + offset.z + well_def.z,
        )

    def get_nominal_well_position(
        self,
        labware_id: str,
        well_name: str,
    ) -> Point:
        """Get the well position without calibration offsets."""
        parent_pos = self.get_labware_parent_nominal_position(labware_id)
        origin_offset = self._labware.get_definition(labware_id).cornerOffsetFromSlot
        well_def = self._labware.get_well_definition(labware_id, well_name)
        return Point(
            x=parent_pos.x + origin_offset.x + well_def.x,
            y=parent_pos.y + origin_offset.y + well_def.y,
            z=parent_pos.z + origin_offset.z + well_def.z + well_def.depth,
        )

    def get_relative_well_location(
        self,
        labware_id: str,
        well_name: str,
        absolute_point: Point,
    ) -> WellLocation:
        """Given absolute position, get relative location of a well in a labware."""
        well_absolute_point = self.get_well_position(labware_id, well_name)
        delta = absolute_point - well_absolute_point

        return WellLocation(offset=WellOffset(x=delta.x, y=delta.y, z=delta.z))

    def get_relative_liquid_handling_well_location(
        self,
        labware_id: str,
        well_name: str,
        absolute_point: Point,
        is_meniscus: Optional[bool] = None,
    ) -> LiquidHandlingWellLocation:
        """Given absolute position, get relative location of a well in a labware.

        If is_meniscus is True, absolute_point will hold the z-offset in its z field.
        """
        if is_meniscus:
            return LiquidHandlingWellLocation(
                origin=WellOrigin.MENISCUS,
                offset=WellOffset(x=0, y=0, z=absolute_point.z),
            )
        else:
            well_absolute_point = self.get_well_position(labware_id, well_name)
            delta = absolute_point - well_absolute_point
            return LiquidHandlingWellLocation(
                offset=WellOffset(x=delta.x, y=delta.y, z=delta.z)
            )

    def get_relative_pick_up_tip_well_location(
        self,
        labware_id: str,
        well_name: str,
        absolute_point: Point,
    ) -> PickUpTipWellLocation:
        """Given absolute position, get relative location of a well in a labware."""
        well_absolute_point = self.get_well_position(labware_id, well_name)
        delta = absolute_point - well_absolute_point

        return PickUpTipWellLocation(offset=WellOffset(x=delta.x, y=delta.y, z=delta.z))

    def get_well_height(
        self,
        labware_id: str,
        well_name: str,
    ) -> float:
        """Get the height of a specified well for a labware."""
        well_def = self._labware.get_well_definition(labware_id, well_name)
        return well_def.depth

    def _get_highest_z_from_labware_data(self, lw_data: LoadedLabware) -> float:
        labware_pos = self.get_labware_position(lw_data.id)
        definition = self._labware.get_definition(lw_data.id)
        z_dim = definition.dimensions.zDimension
        height_over_labware: float = 0
        if isinstance(lw_data.location, ModuleLocation):
            # Note: when calculating highest z of stacked labware, height-over-labware
            # gets accounted for only if the top labware is directly on the module.
            # So if there's a labware on an adapter on a module, then this
            # over-module-height gets ignored. We currently do not have any modules
            # that use an adapter and has height over labware so this doesn't cause
            # any issues yet. But if we add one in the future then this calculation
            # should be updated.
            module_id = lw_data.location.moduleId
            height_over_labware = self._modules.get_height_over_labware(module_id)
        return labware_pos.z + z_dim + height_over_labware

    def get_nominal_effective_tip_length(
        self,
        pipette_id: str,
        labware_id: str,
    ) -> float:
        """Given a labware and a pipette's config, get the nominal effective tip length.

        Effective tip length is the nominal tip length less the distance the
        tip overlaps with the pipette nozzle. This does not take calibrated
        tip lengths into account.
        """
        labware_uri = self._labware.get_definition_uri(labware_id)
        nominal_overlap = self._pipettes.get_nominal_tip_overlap(
            pipette_id=pipette_id, labware_uri=labware_uri
        )

        return self._labware.get_tip_length(
            labware_id=labware_id, overlap=nominal_overlap
        )

    def get_nominal_tip_geometry(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: Optional[str],
    ) -> TipGeometry:
        """Given a labware, well, and hardware pipette config, get the tip geometry.

        Tip geometry includes effective tip length, tip diameter, and tip volume,
        which is all data required by the hardware controller for proper tip handling.

        This geometry data is based solely on labware and pipette definitions and
        does not take calibrated tip lengths into account.
        """
        effective_length = self.get_nominal_effective_tip_length(
            pipette_id=pipette_id,
            labware_id=labware_id,
        )
        well_def = self._labware.get_well_definition(labware_id, well_name)

        if well_def.shape != "circular":
            raise errors.LabwareIsNotTipRackError(
                f"Well {well_name} in labware {labware_id} is not circular."
            )

        return TipGeometry(
            length=effective_length,
            diameter=well_def.diameter,  # type: ignore[arg-type]
            # TODO(mc, 2020-11-12): WellDefinition type says totalLiquidVolume
            #  is a float, but hardware controller expects an int
            volume=int(well_def.totalLiquidVolume),
        )

    def get_checked_tip_drop_location(
        self,
        pipette_id: str,
        labware_id: str,
        well_location: DropTipWellLocation,
        partially_configured: bool = False,
    ) -> WellLocation:
        """Get tip drop location given labware and hardware pipette.

        This makes sure that the well location has an appropriate origin & offset
        if one is not already set previously.
        """
        if (
            self._labware.get_definition(labware_id).parameters.isTiprack
            and partially_configured
        ):
            raise errors.UnexpectedProtocolError(
                "Cannot return tip to a tiprack while the pipette is configured for partial tip."
            )
        if well_location.origin != DropTipWellOrigin.DEFAULT:
            return WellLocation(
                origin=WellOrigin(well_location.origin.value),
                offset=well_location.offset,
            )

        if self._labware.get_definition(labware_id).parameters.isTiprack:
            z_offset = self._labware.get_tip_drop_z_offset(
                labware_id=labware_id,
                length_scale=self._pipettes.get_return_tip_scale(pipette_id),
                additional_offset=well_location.offset.z,
            )
        else:
            # return to top if labware is not tip rack
            z_offset = well_location.offset.z

        return WellLocation(
            origin=WellOrigin.TOP,
            offset=WellOffset(
                x=well_location.offset.x,
                y=well_location.offset.y,
                z=z_offset,
            ),
        )

    def convert_pick_up_tip_well_location(
        self, well_location: PickUpTipWellLocation
    ) -> WellLocation:
        """Convert PickUpTipWellLocation to WellLocation."""
        return WellLocation(
            origin=WellOrigin(well_location.origin.value), offset=well_location.offset
        )

    # TODO(jbl 11-30-2023) fold this function into get_ancestor_slot_name see RSS-411
    def _get_staging_slot_name(self, labware_id: str) -> str:
        """Get the staging slot name that the labware is on."""
        labware_location = self._labware.get(labware_id).location
        if isinstance(labware_location, OnLabwareLocation):
            below_labware_id = labware_location.labwareId
            return self._get_staging_slot_name(below_labware_id)
        elif isinstance(
            labware_location, AddressableAreaLocation
        ) and fixture_validation.is_staging_slot(labware_location.addressableAreaName):
            return labware_location.addressableAreaName
        else:
            raise ValueError(
                "Cannot get staging slot name for labware not on staging slot."
            )

    def _get_lid_dock_slot_name(self, labware_id: str) -> str:
        """Get the staging slot name that the labware is on."""
        labware_location = self._labware.get(labware_id).location
        assert isinstance(labware_location, AddressableAreaLocation)
        return labware_location.addressableAreaName

    def get_ancestor_slot_name(
        self, labware_id: str
    ) -> Union[DeckSlotName, StagingSlotName]:
        """Get the slot name of the labware or the module that the labware is on."""
        labware = self._labware.get(labware_id)
        slot_name: Union[DeckSlotName, StagingSlotName]

        if isinstance(labware.location, DeckSlotLocation):
            slot_name = labware.location.slotName
        elif isinstance(labware.location, ModuleLocation):
            module_id = labware.location.moduleId
            slot_name = self._modules.get_location(module_id).slotName
        elif isinstance(labware.location, OnLabwareLocation):
            below_labware_id = labware.location.labwareId
            slot_name = self.get_ancestor_slot_name(below_labware_id)
        elif isinstance(labware.location, AddressableAreaLocation):
            area_name = labware.location.addressableAreaName
            if self._labware.is_absorbance_reader_lid(labware_id):
                raise errors.LocationIsLidDockSlotError(
                    "Cannot get ancestor slot name for labware on lid dock slot."
                )
            elif fixture_validation.is_staging_slot(area_name):
                slot_name = StagingSlotName.from_primitive(area_name)
            else:
                slot_name = DeckSlotName.from_primitive(area_name)
        elif labware.location == OFF_DECK_LOCATION:
            raise errors.LabwareNotOnDeckError(
                f"Labware {labware_id} does not have a slot associated with it"
                f" since it is no longer on the deck."
            )

        return slot_name

    def ensure_location_not_occupied(
        self, location: _LabwareLocation
    ) -> _LabwareLocation:
        """Ensure that the location does not already have either Labware or a Module in it."""
        # TODO (spp, 2023-11-27): Slot locations can also be addressable areas
        #  so we will need to cross-check against items loaded in both location types.
        #  Something like 'check if an item is in lists of both- labware on addressable areas
        #  as well as labware on slots'. Same for modules.
        if isinstance(
            location,
            (
                DeckSlotLocation,
                ModuleLocation,
                OnLabwareLocation,
                AddressableAreaLocation,
            ),
        ):
            self._labware.raise_if_labware_in_location(location)
        if isinstance(location, DeckSlotLocation):
            self._modules.raise_if_module_in_location(location)
        return location

    def get_labware_grip_point(
        self,
        labware_definition: LabwareDefinition,
        location: Union[
            DeckSlotLocation, ModuleLocation, OnLabwareLocation, AddressableAreaLocation
        ],
    ) -> Point:
        """Get the grip point of the labware as placed on the given location.

        Returns the absolute position of the labware's gripping point as if
        it were placed on the specified location. Labware offset (LPC offset) not included.

        Grip point is the location where critical point of the gripper should move to
        in order to pick/drop the given labware in the specified location.
        It is calculated as the xy center of the slot with z as the point indicated by
        z-position of labware bottom + grip height from labware bottom.
        """
        grip_height_from_labware_bottom = (
            self._labware.get_grip_height_from_labware_bottom(labware_definition)
        )
        location_name: str

        if isinstance(location, DeckSlotLocation):
            location_name = location.slotName.id
            offset = LabwareOffsetVector(x=0, y=0, z=0)
        elif isinstance(location, AddressableAreaLocation):
            location_name = location.addressableAreaName
            if fixture_validation.is_gripper_waste_chute(location_name):
                drop_labware_location = (
                    self._addressable_areas.get_addressable_area_move_to_location(
                        location_name
                    )
                )
                return drop_labware_location + Point(z=grip_height_from_labware_bottom)
            # Location should have been pre-validated so this will be a deck/staging area slot
            else:
                offset = LabwareOffsetVector(x=0, y=0, z=0)
        else:
            if isinstance(location, ModuleLocation):
                location_name = self._modules.get_location(
                    location.moduleId
                ).slotName.id
            else:  # OnLabwareLocation
                location_name = self.get_ancestor_slot_name(location.labwareId).id
            labware_offset = self._get_offset_from_parent(
                child_definition=labware_definition, parent=location
            )
            # Get the calibrated offset if the on labware location is on top of a module, otherwise return empty one
            cal_offset = self._get_calibrated_module_offset(location)
            offset = LabwareOffsetVector(
                x=labware_offset.x + cal_offset.x,
                y=labware_offset.y + cal_offset.y,
                z=labware_offset.z + cal_offset.z,
            )

        location_center = self._addressable_areas.get_addressable_area_center(
            location_name
        )
        return Point(
            location_center.x + offset.x,
            location_center.y + offset.y,
            location_center.z + offset.z + grip_height_from_labware_bottom,
        )

    def get_extra_waypoints(
        self,
        location: Optional[CurrentPipetteLocation],
        to_slot: Union[DeckSlotName, StagingSlotName],
    ) -> List[Tuple[float, float]]:
        """Get extra waypoints for movement if thermocycler needs to be dodged."""
        if location is not None:
            if isinstance(location, CurrentWell):
                from_slot = self.get_ancestor_slot_name(location.labware_id)
            else:
                from_slot = self._addressable_areas.get_addressable_area_base_slot(
                    location.addressable_area_name
                )
            if self._modules.should_dodge_thermocycler(
                from_slot=from_slot, to_slot=to_slot
            ):
                middle_slot = DeckSlotName.SLOT_5.to_equivalent_for_robot_type(
                    self._config.robot_type
                )
                middle_slot_center = (
                    self._addressable_areas.get_addressable_area_center(
                        addressable_area_name=middle_slot.id,
                    )
                )
                return [(middle_slot_center.x, middle_slot_center.y)]
        return []

    def get_slot_item(
        self, slot_name: Union[DeckSlotName, StagingSlotName]
    ) -> Union[LoadedLabware, LoadedModule, CutoutFixture, None]:
        """Get the top-most item present in a deck slot, if any.

        This includes any module that occupies the given slot even if it wasn't loaded
        in that slot (e.g., thermocycler).
        """
        maybe_labware = self._labware.get_by_slot(
            slot_name=slot_name,
        )

        if isinstance(slot_name, DeckSlotName):
            maybe_fixture = self._addressable_areas.get_fixture_by_deck_slot_name(
                slot_name
            )
            # Ignore generic single slot fixtures
            if maybe_fixture and maybe_fixture["id"] in {
                "singleLeftSlot",
                "singleCenterSlot",
                "singleRightSlot",
            }:
                maybe_fixture = None

            maybe_module = self._modules.get_by_slot(
                slot_name=slot_name,
            ) or self._modules.get_overflowed_module_in_slot(slot_name=slot_name)
        else:
            # Modules and fixtures can't be loaded on staging slots
            maybe_fixture = None
            maybe_module = None

        return maybe_labware or maybe_module or maybe_fixture or None

    @staticmethod
    def get_slot_column(slot_name: Union[DeckSlotName, StagingSlotName]) -> int:
        """Get the column number for the specified slot."""
        if isinstance(slot_name, StagingSlotName):
            return 4
        row_col_name = slot_name.to_ot3_equivalent()
        slot_name_match = WELL_NAME_PATTERN.match(row_col_name.value)
        assert (
            slot_name_match is not None
        ), f"Slot name {row_col_name} did not match required pattern; please check labware location."

        row_name, column_name = slot_name_match.group(1, 2)
        return int(column_name)

    def get_next_tip_drop_location(
        self, labware_id: str, well_name: str, pipette_id: str
    ) -> DropTipWellLocation:
        """Get the next location within the specified well to drop the tip into.

        In order to prevent tip stacking, we will alternate between two tip drop locations:
        1. location in left section: a safe distance from left edge of the well
        2. location in right section: a safe distance from right edge of the well

        This safe distance for most cases would be a location where all tips drop
        reliably inside the labware's well. This can be calculated based off of the
        span of a pipette, including all its tips, in the x-direction.

        But we also need to account for the not-so-uncommon case of a left pipette
        trying to drop tips in a labware in the rightmost deck column and vice versa.
        If this labware extends beyond a regular deck slot, like the Flex's default trash,
        then even after keeping a margin for x-span of a pipette, we will get
        a location that's unreachable for the pipette. In such cases, we try to drop tips
        at the rightmost location that a left pipette is able to reach,
        and leftmost location that a right pipette is able to reach respectively.

        In these calculations we assume that the critical point of a pipette
        is considered to be the midpoint of the pipette's tip for single channel,
        and the midpoint of the entire tip assembly for multi-channel pipettes.
        We also assume that the pipette_x_span includes any safety margins required.
        """
        if not self._labware.is_fixed_trash(labware_id=labware_id):
            # In order to avoid the complexity of finding tip drop locations for
            # variety of labware with different well configs, we will allow
            # location cycling only for fixed trash labware right now.
            # TODO (spp, 2023-09-12): update this to possibly a labware-width based check,
            #  or a 'trash' quirk check, once movable trash is implemented.
            return DropTipWellLocation(
                origin=DropTipWellOrigin.DEFAULT,
                offset=WellOffset(x=0, y=0, z=0),
            )

        well_x_dim = self._labware.get_well_size(
            labware_id=labware_id, well_name=well_name
        )[0]
        pipette_channels = self._pipettes.get_config(pipette_id).channels
        pipette_mount = self._pipettes.get_mount(pipette_id)

        labware_slot_column = self.get_slot_column(
            slot_name=self.get_ancestor_slot_name(labware_id)
        )

        if self._last_drop_tip_location_spot.get(labware_id) == _TipDropSection.RIGHT:
            # Drop tip in LEFT section
            x_offset = self._get_drop_tip_well_x_offset(
                tip_drop_section=_TipDropSection.LEFT,
                well_x_dim=well_x_dim,
                pipette_channels=pipette_channels,
                pipette_mount=pipette_mount,
                labware_slot_column=labware_slot_column,
            )
            self._last_drop_tip_location_spot[labware_id] = _TipDropSection.LEFT
        else:
            # Drop tip in RIGHT section
            x_offset = self._get_drop_tip_well_x_offset(
                tip_drop_section=_TipDropSection.RIGHT,
                well_x_dim=well_x_dim,
                pipette_channels=pipette_channels,
                pipette_mount=pipette_mount,
                labware_slot_column=labware_slot_column,
            )
            self._last_drop_tip_location_spot[labware_id] = _TipDropSection.RIGHT

        return DropTipWellLocation(
            origin=DropTipWellOrigin.TOP,
            offset=WellOffset(
                x=x_offset,
                y=0,
                z=0,
            ),
        )

    # TODO find way to combine this with above
    def get_next_tip_drop_location_for_addressable_area(
        self,
        addressable_area_name: str,
        pipette_id: str,
    ) -> AddressableOffsetVector:
        """Get the next location within the specified well to drop the tip into.

        See the doc-string for `get_next_tip_drop_location` for more info on execution.
        """
        area_x_dim = self._addressable_areas.get_addressable_area(
            addressable_area_name
        ).bounding_box.x

        pipette_channels = self._pipettes.get_config(pipette_id).channels
        pipette_mount = self._pipettes.get_mount(pipette_id)

        labware_slot_column = self.get_slot_column(
            slot_name=self._addressable_areas.get_addressable_area_base_slot(
                addressable_area_name
            )
        )

        if (
            self._last_drop_tip_location_spot.get(addressable_area_name)
            == _TipDropSection.RIGHT
        ):
            # Drop tip in LEFT section
            x_offset = self._get_drop_tip_well_x_offset(
                tip_drop_section=_TipDropSection.LEFT,
                well_x_dim=area_x_dim,
                pipette_channels=pipette_channels,
                pipette_mount=pipette_mount,
                labware_slot_column=labware_slot_column,
            )
            self._last_drop_tip_location_spot[
                addressable_area_name
            ] = _TipDropSection.LEFT
        else:
            # Drop tip in RIGHT section
            x_offset = self._get_drop_tip_well_x_offset(
                tip_drop_section=_TipDropSection.RIGHT,
                well_x_dim=area_x_dim,
                pipette_channels=pipette_channels,
                pipette_mount=pipette_mount,
                labware_slot_column=labware_slot_column,
            )
            self._last_drop_tip_location_spot[
                addressable_area_name
            ] = _TipDropSection.RIGHT

        return AddressableOffsetVector(x=x_offset, y=0, z=0)

    @staticmethod
    def _get_drop_tip_well_x_offset(
        tip_drop_section: _TipDropSection,
        well_x_dim: float,
        pipette_channels: int,
        pipette_mount: MountType,
        labware_slot_column: int,
    ) -> float:
        """Get the well x offset for DropTipWellLocation."""
        drop_location_margin_from_labware_edge = (
            PIPETTE_X_SPAN[cast(ChannelCount, pipette_channels)] / 2
        )
        if tip_drop_section == _TipDropSection.LEFT:
            if (
                well_x_dim > SLOT_WIDTH
                and pipette_channels != 96
                and pipette_mount == MountType.RIGHT
                and labware_slot_column == 1
            ):
                # Pipette might not reach the default left spot so use a different left spot
                x_well_offset = (
                    -well_x_dim / 2 + drop_location_margin_from_labware_edge * 2
                )
            else:
                x_well_offset = -well_x_dim / 2 + drop_location_margin_from_labware_edge
                if x_well_offset > 0:
                    x_well_offset = 0
        else:
            if (
                well_x_dim > SLOT_WIDTH
                and pipette_channels != 96
                and pipette_mount == MountType.LEFT
                and labware_slot_column == 3
            ):
                # Pipette might not reach the default right spot so use a different right spot
                x_well_offset = (
                    -well_x_dim / 2
                    + SLOT_WIDTH
                    - drop_location_margin_from_labware_edge
                )
            else:
                x_well_offset = well_x_dim / 2 - drop_location_margin_from_labware_edge
                if x_well_offset < 0:
                    x_well_offset = 0
        return x_well_offset

    def get_final_labware_movement_offset_vectors(
        self,
        from_location: OnDeckLabwareLocation,
        to_location: OnDeckLabwareLocation,
        additional_offset_vector: LabwareMovementOffsetData,
        current_labware: LabwareDefinition,
    ) -> LabwareMovementOffsetData:
        """Calculate the final labware offset vector to use in labware movement."""
        pick_up_offset = (
            self.get_total_nominal_gripper_offset_for_move_type(
                location=from_location,
                move_type=_GripperMoveType.PICK_UP_LABWARE,
                current_labware=current_labware,
            )
            + additional_offset_vector.pickUpOffset
        )
        drop_offset = (
            self.get_total_nominal_gripper_offset_for_move_type(
                location=to_location,
                move_type=_GripperMoveType.DROP_LABWARE,
                current_labware=current_labware,
            )
            + additional_offset_vector.dropOffset
        )

        return LabwareMovementOffsetData(
            pickUpOffset=pick_up_offset, dropOffset=drop_offset
        )

    @staticmethod
    def ensure_valid_gripper_location(
        location: LabwareLocation,
    ) -> Union[
        DeckSlotLocation, ModuleLocation, OnLabwareLocation, AddressableAreaLocation
    ]:
        """Ensure valid on-deck location for gripper, otherwise raise error."""
        if not isinstance(
            location,
            (
                DeckSlotLocation,
                ModuleLocation,
                OnLabwareLocation,
                AddressableAreaLocation,
            ),
        ):
            raise errors.LabwareMovementNotAllowedError(
                "Off-deck labware movements are not supported using the gripper."
            )
        return location

    def get_total_nominal_gripper_offset_for_move_type(
        self,
        location: OnDeckLabwareLocation,
        move_type: _GripperMoveType,
        current_labware: LabwareDefinition,
    ) -> LabwareOffsetVector:
        """Get the total of the offsets to be used to pick up labware in its current location."""
        if move_type == _GripperMoveType.PICK_UP_LABWARE:
            if isinstance(
                location, (ModuleLocation, DeckSlotLocation, AddressableAreaLocation)
            ):
                return self._nominal_gripper_offsets_for_location(location).pickUpOffset
            else:
                # If it's a labware on a labware (most likely an adapter),
                # we calculate the offset as sum of offsets for the direct parent labware
                # and the underlying non-labware parent location.
                direct_parent_offset = self._nominal_gripper_offsets_for_location(
                    location
                )
                ancestor = self._labware.get_parent_location(location.labwareId)
                extra_offset = LabwareOffsetVector(x=0, y=0, z=0)
                if (
                    isinstance(ancestor, ModuleLocation)
                    and self._modules._state.requested_model_by_id[ancestor.moduleId]
                    == ModuleModel.THERMOCYCLER_MODULE_V2
                    and labware_validation.validate_definition_is_lid(current_labware)
                ):
                    if "lidOffsets" in current_labware.gripperOffsets.keys():
                        extra_offset = LabwareOffsetVector(
                            x=current_labware.gripperOffsets[
                                "lidOffsets"
                            ].pickUpOffset.x,
                            y=current_labware.gripperOffsets[
                                "lidOffsets"
                            ].pickUpOffset.y,
                            z=current_labware.gripperOffsets[
                                "lidOffsets"
                            ].pickUpOffset.z,
                        )
                    else:
                        raise errors.LabwareOffsetDoesNotExistError(
                            f"Labware Definition {current_labware.parameters.loadName} does not contain required field 'lidOffsets' of 'gripperOffsets'."
                        )

                assert isinstance(
                    ancestor,
                    (
                        DeckSlotLocation,
                        ModuleLocation,
                        OnLabwareLocation,
                        AddressableAreaLocation,
                    ),
                ), "No gripper offsets for off-deck labware"
                return (
                    direct_parent_offset.pickUpOffset
                    + self._nominal_gripper_offsets_for_location(
                        location=ancestor
                    ).pickUpOffset
                    + extra_offset
                )
        else:
            if isinstance(
                location, (ModuleLocation, DeckSlotLocation, AddressableAreaLocation)
            ):
                return self._nominal_gripper_offsets_for_location(location).dropOffset
            else:
                # If it's a labware on a labware (most likely an adapter),
                # we calculate the offset as sum of offsets for the direct parent labware
                # and the underlying non-labware parent location.
                direct_parent_offset = self._nominal_gripper_offsets_for_location(
                    location
                )
                ancestor = self._labware.get_parent_location(location.labwareId)
                extra_offset = LabwareOffsetVector(x=0, y=0, z=0)
                if (
                    isinstance(ancestor, ModuleLocation)
                    # todo(mm, 2024-11-06): Do not access private module state; only use public ModuleView methods.
                    and self._modules._state.requested_model_by_id[ancestor.moduleId]
                    == ModuleModel.THERMOCYCLER_MODULE_V2
                    and labware_validation.validate_definition_is_lid(current_labware)
                ):
                    if "lidOffsets" in current_labware.gripperOffsets.keys():
                        extra_offset = LabwareOffsetVector(
                            x=current_labware.gripperOffsets[
                                "lidOffsets"
                            ].pickUpOffset.x,
                            y=current_labware.gripperOffsets[
                                "lidOffsets"
                            ].pickUpOffset.y,
                            z=current_labware.gripperOffsets[
                                "lidOffsets"
                            ].pickUpOffset.z,
                        )
                    else:
                        raise errors.LabwareOffsetDoesNotExistError(
                            f"Labware Definition {current_labware.parameters.loadName} does not contain required field 'lidOffsets' of 'gripperOffsets'."
                        )

                assert isinstance(
                    ancestor,
                    (
                        DeckSlotLocation,
                        ModuleLocation,
                        OnLabwareLocation,
                        AddressableAreaLocation,
                    ),
                ), "No gripper offsets for off-deck labware"
                return (
                    direct_parent_offset.dropOffset
                    + self._nominal_gripper_offsets_for_location(
                        location=ancestor
                    ).dropOffset
                    + extra_offset
                )

    # todo(mm, 2024-11-05): This may be incorrect because it does not take the following
    # offsets into account, which *are* taken into account for the actual gripper movement:
    #
    # * The pickup offset in the definition of the parent of the gripped labware.
    # * The "additional offset" or "user offset", e.g. the `pickUpOffset` and `dropOffset`
    #   params in the `moveLabware` command.
    #
    # And this *does* take these extra offsets into account:
    #
    # * The labware's Labware Position Check offset
    #
    # For robustness, we should combine this with `get_gripper_labware_movement_waypoints()`.
    #
    # We should also be more explicit about which offsets act to move the gripper paddles
    # relative to the gripped labware, and which offsets act to change how the gripped
    # labware sits atop its parent. Those have different effects on how far the gripped
    # labware juts beyond the paddles while it's in transit.
    def check_gripper_labware_tip_collision(
        self,
        gripper_homed_position_z: float,
        labware_id: str,
        current_location: OnDeckLabwareLocation,
    ) -> None:
        """Check for potential collision of tips against labware to be lifted."""
        labware_definition = self._labware.get_definition(labware_id)
        pipettes = self._pipettes.get_all()
        for pipette in pipettes:
            # TODO(cb, 2024-01-22): Remove the 1 and 8 channel special case once we are doing X axis validation
            if self._pipettes.get_channels(pipette.id) in [1, 8]:
                return

            tip = self._pipettes.get_attached_tip(pipette.id)
            if tip:
                # NOTE: This call to get_labware_highest_z() uses the labware's LPC offset,
                # which is an inconsistency between this and the actual gripper movement.
                # See the todo comment above this function.
                labware_top_z_when_gripped = gripper_homed_position_z + (
                    self.get_labware_highest_z(labware_id=labware_id)
                    - self.get_labware_grip_point(
                        labware_definition=labware_definition, location=current_location
                    ).z
                )
                # TODO(cb, 2024-01-18): Utilizing the nozzle map and labware X coordinates verify if collisions will occur on the X axis (analysis will use hard coded data to measure from the gripper critical point to the pipette mount)
                if (
                    _PIPETTE_HOMED_POSITION_Z - tip.length
                ) < labware_top_z_when_gripped:
                    raise LabwareMovementNotAllowedError(
                        f"Cannot move labware '{labware_definition.parameters.loadName}' when {int(tip.volume)} µL tips are attached."
                    )
        return

    def _nominal_gripper_offsets_for_location(
        self, location: OnDeckLabwareLocation
    ) -> LabwareMovementOffsetData:
        """Provide the default gripper offset data for the given location type."""
        if isinstance(location, (DeckSlotLocation, AddressableAreaLocation)):
            # TODO we might need a separate type of gripper offset for addressable areas but that also might just
            #   be covered by the drop labware offset/location
            offsets = self._labware.get_deck_default_gripper_offsets()
        elif isinstance(location, ModuleLocation):
            offsets = self._modules.get_default_gripper_offsets(location.moduleId)
        else:
            # Labware is on a labware/adapter
            offsets = self._labware_gripper_offsets(location.labwareId)
        return offsets or LabwareMovementOffsetData(
            pickUpOffset=LabwareOffsetVector(x=0, y=0, z=0),
            dropOffset=LabwareOffsetVector(x=0, y=0, z=0),
        )

    def _labware_gripper_offsets(
        self, labware_id: str
    ) -> Optional[LabwareMovementOffsetData]:
        """Provide the most appropriate gripper offset data for the specified labware.

        We check the types of gripper offsets available for the labware ("default" or slot-based)
        and return the most appropriate one for the overall location of the labware.
        Currently, only module adapters (specifically, the H/S universal flat adapter)
        have non-default offsets that are specific to location of the module on deck,
        so, this code only checks for the presence of those known offsets.
        """
        parent_location = self._labware.get_parent_location(labware_id)
        assert isinstance(
            parent_location,
            (
                DeckSlotLocation,
                ModuleLocation,
                AddressableAreaLocation,
                OnLabwareLocation,
            ),
        ), "No gripper offsets for off-deck labware"

        if isinstance(parent_location, DeckSlotLocation):
            slot_name = parent_location.slotName
        elif isinstance(parent_location, AddressableAreaLocation):
            slot_name = self._addressable_areas.get_addressable_area_base_slot(
                parent_location.addressableAreaName
            )
        else:
            module_loc = self._modules.get_location(parent_location.moduleId)
            slot_name = module_loc.slotName

        slot_based_offset = self._labware.get_child_gripper_offsets(
            labware_id=labware_id, slot_name=slot_name.to_ot3_equivalent()
        )

        return slot_based_offset or self._labware.get_child_gripper_offsets(
            labware_id=labware_id, slot_name=None
        )

    def get_offset_location(self, labware_id: str) -> Optional[LabwareOffsetLocation]:
        """Provide the LabwareOffsetLocation specifying the current position of the labware.

        If the labware is in a location that cannot be specified by a LabwareOffsetLocation
        (for instance, OFF_DECK) then return None.
        """
        parent_location = self._labware.get_location(labware_id)

        if isinstance(parent_location, DeckSlotLocation):
            return LabwareOffsetLocation(
                slotName=parent_location.slotName, moduleModel=None, definitionUri=None
            )
        elif isinstance(parent_location, ModuleLocation):
            module_model = self._modules.get_requested_model(parent_location.moduleId)
            module_location = self._modules.get_location(parent_location.moduleId)
            return LabwareOffsetLocation(
                slotName=module_location.slotName,
                moduleModel=module_model,
                definitionUri=None,
            )
        elif isinstance(parent_location, OnLabwareLocation):
            non_labware_parent_location = self._labware.get_parent_location(labware_id)

            parent_uri = self._labware.get_definition_uri(parent_location.labwareId)
            if isinstance(non_labware_parent_location, DeckSlotLocation):
                return LabwareOffsetLocation(
                    slotName=non_labware_parent_location.slotName,
                    moduleModel=None,
                    definitionUri=parent_uri,
                )
            elif isinstance(non_labware_parent_location, ModuleLocation):
                module_model = self._modules.get_requested_model(
                    non_labware_parent_location.moduleId
                )
                module_location = self._modules.get_location(
                    non_labware_parent_location.moduleId
                )
                return LabwareOffsetLocation(
                    slotName=module_location.slotName,
                    moduleModel=module_model,
                    definitionUri=parent_uri,
                )

        return None

    def get_well_offset_adjustment(
        self,
        labware_id: str,
        well_name: str,
        well_location: WellLocations,
        well_depth: float,
        operation_volume: Optional[float] = None,
    ) -> float:
        """Return a z-axis distance that accounts for well handling height and operation volume.

        Distance is with reference to the well bottom.
        """
        # TODO(pbm, 10-23-24): refactor to smartly reduce height/volume conversions
        initial_handling_height = self.get_well_handling_height(
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            well_depth=well_depth,
        )
        if isinstance(well_location, PickUpTipWellLocation):
            volume = 0.0
        elif isinstance(well_location.volumeOffset, float):
            volume = well_location.volumeOffset
        elif well_location.volumeOffset == "operationVolume":
            volume = operation_volume or 0.0

        if volume:
            return self.get_well_height_after_volume(
                labware_id=labware_id,
                well_name=well_name,
                initial_height=initial_handling_height,
                volume=volume,
            )
        else:
            return initial_handling_height

    def get_meniscus_height(
        self,
        labware_id: str,
        well_name: str,
    ) -> float:
        """Returns stored meniscus height in specified well."""
        well_liquid = self._wells.get_well_liquid_info(
            labware_id=labware_id, well_name=well_name
        )
        if (
            well_liquid.probed_height is not None
            and well_liquid.probed_height.height is not None
        ):
            return well_liquid.probed_height.height
        elif (
            well_liquid.loaded_volume is not None
            and well_liquid.loaded_volume.volume is not None
        ):
            return self.get_well_height_at_volume(
                labware_id=labware_id,
                well_name=well_name,
                volume=well_liquid.loaded_volume.volume,
            )
        elif (
            well_liquid.probed_volume is not None
            and well_liquid.probed_volume.volume is not None
        ):
            return self.get_well_height_at_volume(
                labware_id=labware_id,
                well_name=well_name,
                volume=well_liquid.probed_volume.volume,
            )
        else:
            raise errors.LiquidHeightUnknownError(
                "Must LiquidProbe or LoadLiquid before specifying WellOrigin.MENISCUS."
            )

    def get_well_handling_height(
        self,
        labware_id: str,
        well_name: str,
        well_location: WellLocations,
        well_depth: float,
    ) -> float:
        """Return the handling height for a labware well (with reference to the well bottom)."""
        handling_height = 0.0
        if well_location.origin == WellOrigin.TOP:
            handling_height = well_depth
        elif well_location.origin == WellOrigin.CENTER:
            handling_height = well_depth / 2.0
        elif well_location.origin == WellOrigin.MENISCUS:
            handling_height = self.get_meniscus_height(
                labware_id=labware_id, well_name=well_name
            )
        return float(handling_height)

    def get_well_height_after_volume(
        self, labware_id: str, well_name: str, initial_height: float, volume: float
    ) -> float:
        """Return the height of liquid in a labware well after a given volume has been handled.

        This is given an initial handling height, with reference to the well bottom.
        """
        well_geometry = self._labware.get_well_geometry(
            labware_id=labware_id, well_name=well_name
        )
        initial_volume = find_volume_at_well_height(
            target_height=initial_height, well_geometry=well_geometry
        )
        final_volume = initial_volume + volume
        return find_height_at_well_volume(
            target_volume=final_volume, well_geometry=well_geometry
        )

    def get_well_height_at_volume(
        self, labware_id: str, well_name: str, volume: float
    ) -> float:
        """Convert well volume to height."""
        well_geometry = self._labware.get_well_geometry(labware_id, well_name)
        return find_height_at_well_volume(
            target_volume=volume, well_geometry=well_geometry
        )

    def get_well_volume_at_height(
        self, labware_id: str, well_name: str, height: float
    ) -> float:
        """Convert well height to volume."""
        well_geometry = self._labware.get_well_geometry(labware_id, well_name)
        return find_volume_at_well_height(
            target_height=height, well_geometry=well_geometry
        )

    def validate_dispense_volume_into_well(
        self,
        labware_id: str,
        well_name: str,
        well_location: WellLocations,
        volume: float,
    ) -> None:
        """Raise InvalidDispenseVolumeError if planned dispense volume will overflow well."""
        well_def = self._labware.get_well_definition(labware_id, well_name)
        well_volumetric_capacity = well_def.totalLiquidVolume
        if well_location.origin == WellOrigin.MENISCUS:
            # TODO(pbm, 10-23-24): refactor to smartly reduce height/volume conversions
            well_geometry = self._labware.get_well_geometry(labware_id, well_name)
            meniscus_height = self.get_meniscus_height(
                labware_id=labware_id, well_name=well_name
            )
            meniscus_volume = find_volume_at_well_height(
                target_height=meniscus_height, well_geometry=well_geometry
            )
            remaining_volume = well_volumetric_capacity - meniscus_volume
            if volume > remaining_volume:
                raise errors.InvalidDispenseVolumeError(
                    f"Attempting to dispense {volume}µL of liquid into a well that can currently only hold {remaining_volume}µL (well {well_name} in labware_id: {labware_id})"
                )
        else:
            # TODO(pbm, 10-08-24): factor in well (LabwareStore) state volume
            if volume > well_volumetric_capacity:
                raise errors.InvalidDispenseVolumeError(
                    f"Attempting to dispense {volume}µL of liquid into a well that can only hold {well_volumetric_capacity}µL (well {well_name} in labware_id: {labware_id})"
                )



"""Common pipetting command base models."""
from opentrons_shared_data.errors import ErrorCodes
from pydantic import BaseModel, Field
from typing import Literal, Optional, Tuple, TypedDict

from opentrons.protocol_engine.errors.error_occurrence import ErrorOccurrence

from ..types import WellLocation, LiquidHandlingWellLocation, DeckPoint


class PipetteIdMixin(BaseModel):
    """Mixin for command requests that take a pipette ID."""

    pipetteId: str = Field(
        ...,
        description="Identifier of pipette to use for liquid handling.",
    )


class AspirateVolumeMixin(BaseModel):
    """Mixin for the `volume` field of aspirate commands."""

    volume: float = Field(
        ...,
        description="The amount of liquid to aspirate, in µL."
        " Must not be greater than the remaining available amount, which depends on"
        " the pipette (see `loadPipette`), its configuration (see `configureForVolume`),"
        " the tip (see `pickUpTip`), and the amount you've aspirated so far."
        " There is some tolerance for floating point rounding errors.",
        ge=0,
    )


class DispenseVolumeMixin(BaseModel):
    """Mixin for the `volume` field of dispense commands."""

    volume: float = Field(
        ...,
        description="The amount of liquid to dispense, in µL."
        " Must not be greater than the currently aspirated volume."
        " There is some tolerance for floating point rounding errors.",
        ge=0,
    )


class FlowRateMixin(BaseModel):
    """Mixin for command requests that take a flow rate."""

    flowRate: float = Field(
        ..., description="Speed in µL/s configured for the pipette", gt=0
    )


class WellLocationMixin(BaseModel):
    """Mixin for command requests that take a location that's somewhere in a well."""

    labwareId: str = Field(
        ...,
        description="Identifier of labware to use.",
    )
    wellName: str = Field(
        ...,
        description="Name of well to use in labware.",
    )
    wellLocation: WellLocation = Field(
        default_factory=WellLocation,
        description="Relative well location at which to perform the operation",
    )


class LiquidHandlingWellLocationMixin(BaseModel):
    """Mixin for command requests that take a location that's somewhere in a well."""

    labwareId: str = Field(
        ...,
        description="Identifier of labware to use.",
    )
    wellName: str = Field(
        ...,
        description="Name of well to use in labware.",
    )
    wellLocation: LiquidHandlingWellLocation = Field(
        default_factory=LiquidHandlingWellLocation,
        description="Relative well location at which to perform the operation",
    )


class MovementMixin(BaseModel):
    """Mixin for command requests that move a pipette."""

    minimumZHeight: Optional[float] = Field(
        None,
        description=(
            "Optional minimal Z margin in mm."
            " If this is larger than the API's default safe Z margin,"
            " it will make the arc higher. If it's smaller, it will have no effect."
        ),
    )

    forceDirect: bool = Field(
        False,
        description=(
            "If true, moving from one labware/well to another"
            " will not arc to the default safe z,"
            " but instead will move directly to the specified location."
            " This will also force the `minimumZHeight` param to be ignored."
            " A 'direct' movement is in X/Y/Z simultaneously."
        ),
    )

    speed: Optional[float] = Field(
        None,
        description=(
            "Override the travel speed in mm/s."
            " This controls the straight linear speed of motion."
        ),
    )


class BaseLiquidHandlingResult(BaseModel):
    """Base properties of a liquid handling result."""

    volume: float = Field(
        ...,
        description="Amount of liquid in uL handled in the operation.",
        ge=0,
    )


class DestinationPositionResult(BaseModel):
    """Mixin for command results that move a pipette."""

    # todo(mm, 2024-08-02): Consider deprecating or redefining this.
    #
    # This is here because opentrons.protocol_engine needed it for internal bookkeeping
    # and, at the time, we didn't have a way to do that without adding this to the
    # public command results. Its usefulness to callers outside
    # opentrons.protocol_engine is questionable because they would need to know which
    # critical point is in play, and I think that can change depending on obscure
    # things like labware quirks.
    position: DeckPoint = Field(
        DeckPoint(x=0, y=0, z=0),
        description=(
            "The (x,y,z) coordinates of the pipette's critical point in deck space"
            " after the move was completed."
        ),
    )


class ErrorLocationInfo(TypedDict):
    """Holds a retry location for in-place error recovery.

    This is appropriate to pass to a `moveToCoordinates` command,
    assuming the pipette has not been configured with a different nozzle layout
    in the meantime.
    """

    retryLocation: Tuple[float, float, float]


class OverpressureError(ErrorOccurrence):
    """Returned when sensors detect an overpressure error while moving liquid.

    The pipette plunger motion is stopped at the point of the error.

    The next thing to move the plunger must account for the robot not having a valid
    estimate of its position. It should be a `home`, `unsafe/updatePositionEstimators`,
    `unsafe/dropTipInPlace`, or `unsafe/blowOutInPlace`.
    """

    isDefined: bool = True

    errorType: Literal["overpressure"] = "overpressure"

    errorCode: str = ErrorCodes.PIPETTE_OVERPRESSURE.value.code
    detail: str = ErrorCodes.PIPETTE_OVERPRESSURE.value.detail

    errorInfo: ErrorLocationInfo


class LiquidNotFoundError(ErrorOccurrence):
    """Returned when no liquid is detected during the liquid probe process/move.

    After a failed probing, the pipette returns to the process start position.
    """

    isDefined: bool = True

    errorType: Literal["liquidNotFound"] = "liquidNotFound"

    errorCode: str = ErrorCodes.PIPETTE_LIQUID_NOT_FOUND.value.code
    detail: str = ErrorCodes.PIPETTE_LIQUID_NOT_FOUND.value.detail


class TipPhysicallyAttachedError(ErrorOccurrence):
    """Returned when sensors determine that a tip remains on the pipette after a drop attempt.

    The pipette will act as if the tip was not dropped. So, you won't be able to pick
    up a new tip without dropping the current one, and movement commands will assume
    there is a tip hanging off the bottom of the pipette.
    """

    isDefined: bool = True

    errorType: Literal["tipPhysicallyAttached"] = "tipPhysicallyAttached"

    errorCode: str = ErrorCodes.TIP_DROP_FAILED.value.code
    detail: str = ErrorCodes.TIP_DROP_FAILED.value.detail

    errorInfo: ErrorLocationInfo



from __future__ import annotations
import asyncio
import logging
import re
from typing import TYPE_CHECKING, List, Optional, Union
from glob import glob

from opentrons.config import IS_ROBOT, IS_LINUX
from opentrons.drivers.rpi_drivers import types, interfaces, usb, usb_simulator
from opentrons.hardware_control.emulation.module_server.helpers import (
    listen_module_connection,
)
from opentrons.hardware_control.modules.absorbance_reader import AbsorbanceReader
from opentrons.hardware_control.modules.module_calibration import (
    ModuleCalibrationOffset,
    load_module_calibration_offset,
    save_module_calibration_offset,
)
from opentrons.hardware_control.modules.types import ModuleAtPort, ModuleType
from opentrons.hardware_control.modules import SimulatingModuleAtPort

from opentrons.types import Point
from .types import AionotifyEvent, BoardRevision, OT3Mount
from . import modules

if TYPE_CHECKING:
    from .api import API
    from .ot3api import OT3API


log = logging.getLogger(__name__)

MODULE_PORT_REGEX = re.compile(
    # add a negative lookbehind to suppress matches on OT-2 tempfiles udev creates
    r"(?<!\.#ot_module_)"
    # capture all modules by name using alternation
    + "(" + "|".join(modules.MODULE_TYPE_BY_NAME.keys()) + ")"
    # add a negative lookahead to suppress matches on Flex tempfiles udev creates
    + r"\d+(?!\.tmp-c\d+:\d+)",
    re.I,
)


class AttachedModulesControl:
    """
    A class to handle monitoring module attachment, capturing the physical
    USB port information and finally building a module object.
    """

    def __init__(
        self,
        api: Union["API", "OT3API"],
        usb: interfaces.USBDriverInterface,
    ) -> None:
        self._available_modules: List[modules.AbstractModule] = []
        self._api = api
        self._usb = usb

    @classmethod
    async def build(
        cls,
        api_instance: Union["API", "OT3API"],
        board_revision: BoardRevision,
    ) -> AttachedModulesControl:
        usb_instance = (
            usb.USBBus(board_revision)
            if not api_instance.is_simulator and IS_ROBOT
            else usb_simulator.USBBusSimulator()
        )
        mc_instance = cls(api=api_instance, usb=usb_instance)

        if not api_instance.is_simulator:
            # Do an initial scan of modules.
            await mc_instance.register_modules(mc_instance.scan())
            if not IS_ROBOT:
                # Start task that registers emulated modules.
                api_instance.loop.create_task(
                    listen_module_connection(mc_instance.register_modules)
                )

        return mc_instance

    @property
    def available_modules(self) -> List[modules.AbstractModule]:
        return self._available_modules

    async def build_module(
        self,
        port: str,
        usb_port: types.USBPort,
        type: modules.ModuleType,
        sim_model: Optional[str] = None,
        sim_serial_number: Optional[str] = None,
    ) -> modules.AbstractModule:
        return await modules.build(
            port=port,
            usb_port=usb_port,
            type=type,
            simulating=self._api.is_simulator,
            hw_control_loop=self._api.loop,
            execution_manager=self._api._execution_manager,
            sim_model=sim_model,
            sim_serial_number=sim_serial_number,
            disconnected_callback=self._disconnected_callback,
        )

    def _disconnected_callback(self, port: str, serial: Optional[str]) -> None:
        """Used by the module to indicate that it was disconnected and should be deleted."""
        mod = ModuleAtPort(port=port, serial=serial, name="")
        asyncio.run_coroutine_threadsafe(
            self.unregister_modules([mod]),
            self._api.loop,
        )

    async def unregister_modules(
        self,
        mods_at_ports: Union[
            List[modules.ModuleAtPort], List[modules.SimulatingModuleAtPort]
        ],
    ) -> None:
        """
        De-register Modules.

        Remove any modules that are no longer found by aionotify.
        """
        removed_modules = []
        for mod in mods_at_ports:
            for attached_mod in self.available_modules:
                if (
                    attached_mod.serial_number == mod.serial
                    or attached_mod.port == mod.port
                ):
                    removed_modules.append(attached_mod)

        for removed_mod in removed_modules:
            try:
                self._available_modules.remove(removed_mod)
            except ValueError:
                log.exception(
                    f"Removed Module {removed_mod} not found in attached modules"
                )
        for removed_mod in removed_modules:
            log.info(
                f"Module {removed_mod.name()} detached from port {removed_mod.port}"
            )
            await removed_mod.cleanup()
        self._available_modules = sorted(
            self._available_modules, key=modules.AbstractModule.sort_key
        )

    async def register_modules(
        self,
        new_mods_at_ports: Optional[
            Union[List[modules.ModuleAtPort], List[modules.SimulatingModuleAtPort]]
        ] = None,
        removed_mods_at_ports: Optional[List[modules.ModuleAtPort]] = None,
    ) -> None:
        """
        Register Modules.

        Upon system recognition of a module being plugged in, we should
        register that module and de-register any modules that are
        no longer found on the system.
        """
        if new_mods_at_ports is None:
            new_mods_at_ports = []
        if removed_mods_at_ports is None:
            removed_mods_at_ports = []

        # destroy removed mods
        await self.unregister_modules(removed_mods_at_ports)
        self._new_mods_at_ports = new_mods_at_ports
        unsorted_mods_at_port = self._usb.match_virtual_ports(new_mods_at_ports)

        # build new mods
        for mod in unsorted_mods_at_port:
            new_instance = await self.build_module(
                port=mod.port,
                usb_port=mod.usb_port,
                type=modules.MODULE_TYPE_BY_NAME[mod.name],
                sim_serial_number=(
                    mod.serial_number
                    if isinstance(mod, SimulatingModuleAtPort)
                    else None
                ),
                sim_model=(
                    mod.model if isinstance(mod, SimulatingModuleAtPort) else None
                ),
            )
            self._available_modules.append(new_instance)
            log.info(
                f"Module {mod.name} discovered and attached"
                f" at port {mod.port}, new_instance: {new_instance}"
            )
        self._available_modules = sorted(
            self._available_modules, key=modules.AbstractModule.sort_key
        )

    def scan(self) -> List[modules.ModuleAtPort]:
        """Scan for connected modules and return list of
        tuples of serial ports and device names
        """
        if IS_ROBOT and IS_LINUX:
            devices = glob("/dev/ot_module*")
        else:
            devices = []

        discovered_modules = []

        for port in devices:
            symlink_port = port.split("dev/")[1]
            module_at_port = self.get_module_at_port(symlink_port)
            if module_at_port:
                discovered_modules.append(module_at_port)

        log.debug("Discovered modules: {}".format(discovered_modules))
        return discovered_modules

    @staticmethod
    def get_module_at_port(port: str) -> Optional[modules.ModuleAtPort]:
        """Given a port, returns either a ModuleAtPort
        if it is a recognized module, or None if not recognized.
        """
        match = MODULE_PORT_REGEX.search(port)
        if match:
            name = match.group(1).lower()
            if name not in modules.MODULE_TYPE_BY_NAME:
                log.warning(f"Unexpected module connected: {name} on {port}")
                return None
            return modules.ModuleAtPort(port=f"/dev/{port}", name=name)
        return None

    async def handle_module_appearance(self, event: AionotifyEvent) -> None:
        """Only called upon availability of aionotify. Check that
        the file system has changed and either remove or add modules
        depending on the result.

        Args:
            event: The event passed from aionotify.

        Returns:
            None
        """
        maybe_module_at_port = self.get_module_at_port(event.name)
        new_modules = None
        removed_modules = None
        if maybe_module_at_port is not None:
            if hasattr(event.flags, "DELETE") or hasattr(event.flags, "MOVED_FROM"):
                # NOTE: The absorbance reader is a hidraw device, when we first
                # plug it into the Flex, udev rules create a
                # /dev/ot_module_absorbancereader(n) symlink which aionotify
                # detects as a CREATE event and registers an absorbance module.
                # When we create this Absorbance module and connect to it,
                # the Byonoy library opens the device via hidapi which removes
                # the symlink and triggers a DELETE action from aionoify.
                # When the device is deleted, we disconnect from it causing
                # the /dev/ot_module_absorbance(n) symlink to get created, repeating
                # the cycle.

                # This DELETE action would normally delete the device, but in this case
                # Lets ignore these events for the absorbance reader and handle
                # cleanup when the poller attempts to read data and fails.
                if maybe_module_at_port.name == "absorbancereader":
                    return
                removed_modules = [maybe_module_at_port]
                log.info(f"Module Removed: {maybe_module_at_port}")
            elif hasattr(event.flags, "CREATE") or hasattr(event.flags, "MOVED_TO"):
                # NOTE: Absorbance reader gets disonnected when updating, so
                # if the device is updating, dont create a new instance.
                for module in self._available_modules:
                    if (
                        maybe_module_at_port.name == "absorbancereader"
                        and isinstance(module, AbsorbanceReader)
                        and module.updating
                    ):
                        return

                new_modules = [maybe_module_at_port]
                log.info(f"Module Added: {maybe_module_at_port}")
            try:
                await self.register_modules(
                    removed_mods_at_ports=removed_modules,
                    new_mods_at_ports=new_modules,
                )
            except Exception:
                log.exception("Exception in Module registration")

    def get_module_by_module_id(
        self, module_id: str
    ) -> Optional[modules.AbstractModule]:
        """Returns the module with the matching serial id."""
        found_module: Optional[modules.AbstractModule] = None
        for module in self.available_modules:
            if module.device_info["serial"] == module_id:
                found_module = module
                break
        return found_module

    def load_module_offset(
        self, module_type: ModuleType, module_id: str
    ) -> Optional[ModuleCalibrationOffset]:
        log.info(f"Loading module offset for {module_type} {module_id}")
        return load_module_calibration_offset(module_type, module_id)

    def save_module_offset(
        self,
        module: ModuleType,
        module_id: str,
        mount: OT3Mount,
        slot: str,
        offset: Point,
        instrument_id: Optional[str] = None,
    ) -> Optional[ModuleCalibrationOffset]:
        log.info(f"Saving module {module} {module_id} offset: {offset} for slot {slot}")
        save_module_calibration_offset(
            offset, mount, slot, module, module_id, instrument_id
        )
        return load_module_calibration_offset(module, module_id)



"""Command models to read absorbance."""
from __future__ import annotations
from datetime import datetime
from typing import Optional, Dict, TYPE_CHECKING, List
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors import CannotPerformModuleAction, StorageLimitReachedError
from ...errors.error_occurrence import ErrorOccurrence

from ...resources.file_provider import (
    PlateReaderData,
    ReadData,
    MAXIMUM_CSV_FILE_LIMIT,
)
from ...resources import FileProvider

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler


ReadAbsorbanceCommandType = Literal["absorbanceReader/read"]


class ReadAbsorbanceParams(BaseModel):
    """Input parameters for an absorbance reading."""

    moduleId: str = Field(..., description="Unique ID of the Absorbance Reader.")
    fileName: Optional[str] = Field(
        None,
        description="Optional file name to use when storing the results of a measurement.",
    )


class ReadAbsorbanceResult(BaseModel):
    """Result data from running an aborbance reading, returned as a dictionary map of wavelengths containing a map of values by well name (eg. {450: {"A1": 0.0, ...}})."""

    data: Optional[Dict[int, Dict[str, float]]] = Field(
        ..., description="Absorbance data points per wavelength."
    )
    fileIds: Optional[List[str]] = Field(
        ...,
        description="List of file IDs for files output as a result of a Read action.",
    )


class ReadAbsorbanceImpl(
    AbstractCommandImpl[ReadAbsorbanceParams, SuccessData[ReadAbsorbanceResult]]
):
    """Execution implementation of an Absorbance Reader measurement."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        file_provider: FileProvider,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._file_provider = file_provider

    async def execute(  # noqa: C901
        self, params: ReadAbsorbanceParams
    ) -> SuccessData[ReadAbsorbanceResult]:
        """Initiate an absorbance measurement."""
        abs_reader_substate = self._state_view.modules.get_absorbance_reader_substate(
            module_id=params.moduleId
        )

        # Allow propagation of ModuleNotAttachedError.
        abs_reader = self._equipment.get_module_hardware_api(
            abs_reader_substate.module_id
        )

        if abs_reader_substate.configured is False:
            raise CannotPerformModuleAction(
                "Cannot perform Read action on Absorbance Reader without calling `.initialize(...)` first."
            )
        if abs_reader_substate.is_lid_on is False:
            raise CannotPerformModuleAction(
                "Absorbance Plate Reader can't read a plate with the lid open. Call `close_lid()` first."
            )

        # TODO: we need to return a file ID and increase the file count even when a moduel is not attached
        if (
            params.fileName is not None
            and abs_reader_substate.configured_wavelengths is not None
        ):
            # Validate that the amount of files we are about to generate does not put us higher than the limit
            if (
                self._state_view.files.get_filecount()
                + len(abs_reader_substate.configured_wavelengths)
                > MAXIMUM_CSV_FILE_LIMIT
            ):
                raise StorageLimitReachedError(
                    message=f"Attempt to write file {params.fileName} exceeds file creation limit of {MAXIMUM_CSV_FILE_LIMIT} files."
                )

        asbsorbance_result: Dict[int, Dict[str, float]] = {}
        transform_results = []
        # Handle the measurement and begin building data for return
        if abs_reader is not None:
            start_time = datetime.now()
            results = await abs_reader.start_measure()
            finish_time = datetime.now()
            if abs_reader._measurement_config is not None:
                sample_wavelengths = abs_reader._measurement_config.sample_wavelengths
                for wavelength, result in zip(sample_wavelengths, results):
                    converted_values = (
                        self._state_view.modules.convert_absorbance_reader_data_points(
                            data=result
                        )
                    )
                    asbsorbance_result[wavelength] = converted_values
                    transform_results.append(
                        ReadData.construct(wavelength=wavelength, data=converted_values)
                    )
        # Handle the virtual module case for data creation (all zeroes)
        elif self._state_view.config.use_virtual_modules:
            start_time = finish_time = datetime.now()
            if abs_reader_substate.configured_wavelengths is not None:
                for wavelength in abs_reader_substate.configured_wavelengths:
                    converted_values = (
                        self._state_view.modules.convert_absorbance_reader_data_points(
                            data=[0] * 96
                        )
                    )
                    asbsorbance_result[wavelength] = converted_values
                    transform_results.append(
                        ReadData.construct(wavelength=wavelength, data=converted_values)
                    )
            else:
                raise CannotPerformModuleAction(
                    "Plate Reader data cannot be requested with a module that has not been initialized."
                )

        # TODO (cb, 10-17-2024): FILE PROVIDER - Some day we may want to break the file provider behavior into a seperate API function.
        # When this happens, we probably will to have the change the command results handler we utilize to track file IDs in engine.
        # Today, the action handler for the FileStore looks for a ReadAbsorbanceResult command action, this will need to be delinked.

        # Begin interfacing with the file provider if the user provided a filename
        file_ids = []
        if params.fileName is not None:
            # Create the Plate Reader Transform
            plate_read_result = PlateReaderData.construct(
                read_results=transform_results,
                reference_wavelength=abs_reader_substate.reference_wavelength,
                start_time=start_time,
                finish_time=finish_time,
                serial_number=abs_reader.serial_number
                if (abs_reader is not None and abs_reader.serial_number is not None)
                else "VIRTUAL_SERIAL",
            )

            if isinstance(plate_read_result, PlateReaderData):
                # Write a CSV file for each of the measurements taken
                for measurement in plate_read_result.read_results:
                    file_id = await self._file_provider.write_csv(
                        write_data=plate_read_result.build_generic_csv(
                            filename=params.fileName,
                            measurement=measurement,
                        )
                    )
                    file_ids.append(file_id)

                # Return success data to api
                return SuccessData(
                    public=ReadAbsorbanceResult(
                        data=asbsorbance_result, fileIds=file_ids
                    ),
                )

        return SuccessData(
            public=ReadAbsorbanceResult(data=asbsorbance_result, fileIds=file_ids),
        )


class ReadAbsorbance(
    BaseCommand[ReadAbsorbanceParams, ReadAbsorbanceResult, ErrorOccurrence]
):
    """A command to execute an Absorbance Reader measurement."""

    commandType: ReadAbsorbanceCommandType = "absorbanceReader/read"
    params: ReadAbsorbanceParams
    result: Optional[ReadAbsorbanceResult]

    _ImplementationCls: Type[ReadAbsorbanceImpl] = ReadAbsorbanceImpl


class ReadAbsorbanceCreate(BaseCommandCreate[ReadAbsorbanceParams]):
    """A request to execute an Absorbance Reader measurement."""

    commandType: ReadAbsorbanceCommandType = "absorbanceReader/read"
    params: ReadAbsorbanceParams

    _CommandCls: Type[ReadAbsorbance] = ReadAbsorbance



"""Wait for duration command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from ..execution import RunControlHandler


WaitForDurationCommandType = Literal["waitForDuration"]


class WaitForDurationParams(BaseModel):
    """Payload required to pause the protocol."""

    seconds: float = Field(..., description="Duration, in seconds, to wait for.")
    message: Optional[str] = Field(
        None,
        description="A user-facing message associated with the pause",
    )


class WaitForDurationResult(BaseModel):
    """Result data from the execution of a wait for duration command."""


class WaitForDurationImplementation(
    AbstractCommandImpl[WaitForDurationParams, SuccessData[WaitForDurationResult]]
):
    """Wait for duration command implementation."""

    def __init__(self, run_control: RunControlHandler, **kwargs: object) -> None:
        self._run_control = run_control

    async def execute(
        self, params: WaitForDurationParams
    ) -> SuccessData[WaitForDurationResult]:
        """Wait for a duration of time."""
        await self._run_control.wait_for_duration(params.seconds)
        return SuccessData(
            public=WaitForDurationResult(),
        )


class WaitForDuration(
    BaseCommand[WaitForDurationParams, WaitForDurationResult, ErrorOccurrence]
):
    """Wait for duration command model."""

    commandType: WaitForDurationCommandType = "waitForDuration"
    params: WaitForDurationParams
    result: Optional[WaitForDurationResult]

    _ImplementationCls: Type[
        WaitForDurationImplementation
    ] = WaitForDurationImplementation


class WaitForDurationCreate(BaseCommandCreate[WaitForDurationParams]):
    """Wait for duration command request model."""

    commandType: WaitForDurationCommandType = "waitForDuration"
    params: WaitForDurationParams

    _CommandCls: Type[WaitForDuration] = WaitForDuration



from opentrons.hardware_control.types import StatusBarState
from opentrons_hardware.hardware_control import status_bar
from opentrons_hardware.firmware_bindings.binary_constants import (
    LightAnimationType,
    LightTransitionType,
)

from typing import List


class StatusBarStateController:
    """Stateful control of the status bar.

    Note that the controller can be enabled/disabled at any time. If the controller
    is ever disabled, the lights are set to OFF and future state changes will not
    write to the rear panel board. However, the proper `StatusBarState` will still
    be cached, so if the controller becomes enabled/disabled in the future it will
    be able to resume the correct animation based on the current system state."""

    def __init__(self, controller: status_bar.StatusBar) -> None:
        """Create a StatusBarStateController."""
        self._status_bar_state = StatusBarState.IDLE
        self._controller = controller
        self._enabled = True

    async def _status_bar_idle(self) -> None:
        self._status_bar_state = StatusBarState.IDLE
        if self._enabled:
            await self._controller.static_color(status_bar.WHITE)

    async def _status_bar_running(self) -> None:
        self._status_bar_state = StatusBarState.RUNNING
        if self._enabled:
            await self._controller.static_color(status_bar.GREEN)

    async def _status_bar_paused(self) -> None:
        self._status_bar_state = StatusBarState.PAUSED
        if self._enabled:
            await self._controller.pulse_color(status_bar.BLUE)

    async def _status_bar_hardware_error(self) -> None:
        self._status_bar_state = StatusBarState.HARDWARE_ERROR
        if self._enabled:
            await self._controller.flash_color(status_bar.RED)

    async def _status_bar_software_error(self) -> None:
        self._status_bar_state = StatusBarState.SOFTWARE_ERROR
        if self._enabled:
            await self._controller.static_color(status_bar.YELLOW)

    async def _status_bar_error_recovery(self) -> None:
        self._status_bar_state = StatusBarState.ERROR_RECOVERY
        if self._enabled:
            await self._controller.pulse_color(status_bar.YELLOW)

    async def _status_bar_confirm(self) -> None:
        # Confirm should revert to IDLE
        self._status_bar_state = StatusBarState.IDLE
        if self._enabled:
            await self._controller.blink_once(status_bar.GREEN, status_bar.WHITE)

    async def _status_bar_run_complete(self) -> None:
        self._status_bar_state = StatusBarState.RUN_COMPLETED
        if self._enabled:
            await self._controller.pulse_color(status_bar.GREEN)

    async def _status_bar_updating(self) -> None:
        self._status_bar_state = StatusBarState.UPDATING
        if self._enabled:
            await self._controller.pulse_color(status_bar.WHITE)

    async def _status_bar_activation(self) -> None:
        # Activation should revert to IDLE
        self._status_bar_state = StatusBarState.IDLE

        # This animation uses an intermediate color between the blue and the white.
        # This results in a sort of light-blue effect.
        steps: List[status_bar.ColorStep] = [
            status_bar.ColorStep(LightTransitionType.linear, 250, status_bar.OFF),
            status_bar.ColorStep(LightTransitionType.linear, 750, status_bar.OFF),
            status_bar.ColorStep(LightTransitionType.linear, 1000, status_bar.BLUE),
            status_bar.ColorStep(
                LightTransitionType.linear,
                250,
                status_bar.Color(
                    r=0,
                    g=int(status_bar.BLUE.g * 0.75),
                    b=int(status_bar.BLUE.b * 0.75),
                    w=50,
                ),
            ),
            status_bar.ColorStep(LightTransitionType.linear, 250, status_bar.WHITE),
        ]

        if self._enabled:
            await self._controller.start_animation(
                steps=steps, type=LightAnimationType.single_shot
            )

    async def _status_bar_disco(self) -> None:
        # TODO - update implementation
        colors = [
            status_bar.GREEN,
            status_bar.YELLOW,
            status_bar.PURPLE,
            status_bar.ORANGE,
            status_bar.GREEN,
            status_bar.YELLOW,
            status_bar.PURPLE,
            status_bar.ORANGE,
            status_bar.WHITE,
        ]
        steps: List[status_bar.ColorStep] = [
            status_bar.ColorStep(LightTransitionType.linear, 300, color)
            for color in colors
        ]
        steps.append(
            status_bar.ColorStep(LightTransitionType.linear, 500, status_bar.OFF)
        )
        steps.append(
            status_bar.ColorStep(LightTransitionType.linear, 75, status_bar.BLUE)
        )
        steps.append(
            status_bar.ColorStep(
                LightTransitionType.instant,
                transition_time_ms=150,
                color=status_bar.GREEN,
            )
        )
        steps.append(
            status_bar.ColorStep(
                LightTransitionType.instant,
                transition_time_ms=150,
                color=status_bar.OFF,
            )
        )
        steps.append(
            status_bar.ColorStep(
                LightTransitionType.instant,
                transition_time_ms=150,
                color=status_bar.GREEN,
            )
        )
        steps.append(
            status_bar.ColorStep(
                LightTransitionType.linear,
                transition_time_ms=1000,
                color=status_bar.WHITE,
            )
        )

        self._status_bar_state = StatusBarState.IDLE
        if self._enabled:
            await self._controller.start_animation(
                steps=steps, type=LightAnimationType.single_shot
            )

    async def _status_bar_off(self) -> None:
        self._status_bar_state = StatusBarState.OFF
        if self._enabled:
            await self._controller.static_color(status_bar.OFF)

    async def set_status_bar_state(self, state: StatusBarState) -> None:
        """Main interface to set a new state."""
        callbacks = {
            StatusBarState.IDLE: self._status_bar_idle,
            StatusBarState.RUNNING: self._status_bar_running,
            StatusBarState.PAUSED: self._status_bar_paused,
            StatusBarState.HARDWARE_ERROR: self._status_bar_hardware_error,
            StatusBarState.SOFTWARE_ERROR: self._status_bar_software_error,
            StatusBarState.ERROR_RECOVERY: self._status_bar_error_recovery,
            StatusBarState.CONFIRMATION: self._status_bar_confirm,
            StatusBarState.RUN_COMPLETED: self._status_bar_run_complete,
            StatusBarState.UPDATING: self._status_bar_updating,
            StatusBarState.ACTIVATION: self._status_bar_activation,
            StatusBarState.DISCO: self._status_bar_disco,
            StatusBarState.OFF: self._status_bar_off,
        }
        await callbacks[state]()

    def get_current_state(self) -> StatusBarState:
        """Get the current state."""
        return self._status_bar_state

    async def set_enabled(self, enabled: bool) -> None:
        """Enable or disable the status bar.

        If the status bar is newly disabled, the lights are set OFF
        and the old status is cached so it can later be restored.

        If the status bar is newly enabled, the last-set animation
        will be sent to the rear panel.

        NOTE When the status bar is disabled, `set_status_bar_state` will
        still update the internal status bar setting as if the bar is on!
        This means that, when the status bar is reenabled, it will look
        correct for the current robot status."""
        if enabled == self._enabled:
            return

        if enabled:
            # Need to turn Enabled true *first* so that we can actually write
            # the setting to the rear panel
            self._enabled = True
            await self.set_status_bar_state(self._status_bar_state)
        else:
            cached_state = self.get_current_state()
            await self.set_status_bar_state(StatusBarState.OFF)
            self._enabled = False
            # This ensures that the status bar will be set to the correct state
            # if it's reactivated before the state gets set again.
            self._status_bar_state = cached_state

    def get_enabled(self) -> bool:
        """Check whether the status bar is enabled."""
        return self._enabled



"""
opentrons.system.log_control: functions for talking to syslog and journald

This is the implementation of the endpoints in
:py:mod:`opentrons.server.endpoints.logs` and friends.
"""
import asyncio
import logging
import subprocess
from typing import List


LOG = logging.getLogger(__name__)

MAX_RECORDS = 100000
DEFAULT_RECORDS = 50000

UNIT_SELECTORS = [
    "opentrons-robot-server",
    "opentrons-update-server",
    "opentrons-robot-app",
]
SERIAL_SPECIAL = "ALL_SERIAL"
SERIAL_SELECTORS = [
    "opentrons-api-serial",
    "opentrons-api-serial-can",
    "opentrons-api-serial-usbbin",
]


async def get_records_dumb(selector: str, records: int, mode: str) -> bytes:
    """Dump the log files.

    :param selector: The syslog selector to limit responses to
    :param records: The maximum number of records to print
    :param mode: A journalctl dump mode. Should be either "short-precise" or "json".
    """
    selector_array: List[str] = []
    if selector == SERIAL_SPECIAL:
        for serial_selector in SERIAL_SELECTORS:
            selector_array.extend(["-t", serial_selector])
    elif selector in UNIT_SELECTORS:
        selector_array.extend(["-u", selector])
    else:
        selector_array.extend(["-t", selector])

    proc = await asyncio.create_subprocess_exec(
        "journalctl",
        "--no-pager",
        *selector_array,
        "-n",
        str(records),
        "-o",
        mode,
        "-a",
        stdout=subprocess.PIPE,
    )
    stdout, _ = await proc.communicate()
    return stdout



from typing_extensions import Protocol


class Stoppable(Protocol):
    """Protocol specifying controllability of teardown"""

    async def clean_up(self) -> None:
        """Get the API ready to stop cleanly."""
        ...
